# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST API for held subscription requests."""

from mailman.app.moderator import send_rejection
from mailman.core.i18n import _
from mailman.interfaces.action import Action
from mailman.interfaces.member import AlreadySubscribedError
from mailman.interfaces.pending import IPendings, PendType
from mailman.interfaces.subscriptions import ISubscriptionManager, TokenOwner
from mailman.rest.helpers import (
    bad_request,
    child,
    CollectionMixin,
    conflict,
    etag,
    no_content,
    not_found,
    okay,
)
from mailman.rest.validator import enum_validator, Validator
from public import public
from zope.component import getUtility


class _ModerationBase:
    """Common base class."""

    def __init__(self):
        self._pendings = getUtility(IPendings)

    def _resource_as_dict(self, token):
        pendable = self._pendings.confirm(token, expunge=False)
        if pendable is None:
            # This token isn't in the database.
            raise LookupError
        resource = dict(token=token)
        resource.update(pendable)
        return resource


@public
class IndividualRequest(_ModerationBase):
    """Resource for moderating a membership change."""

    def __init__(self, mlist, token):
        super().__init__()
        self._mlist = mlist
        self._registrar = ISubscriptionManager(self._mlist)
        self._token = token

    def on_get(self, request, response):
        # Get the pended record associated with this token, if it exists in
        # the pending table.
        try:
            resource = self._resource_as_dict(self._token)
            assert resource is not None, resource
        except LookupError:
            not_found(response)
            return
        okay(response, etag(resource))

    def on_post(self, request, response):
        try:
            validator = Validator(action=enum_validator(Action),
                                  reason=str,
                                  _optional=('reason',))
            arguments = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
            return
        action = arguments['action']
        if action in (Action.defer, Action.hold):
            # At least see if the token is in the database.
            pendable = self._pendings.confirm(self._token, expunge=False)
            if pendable is None:
                not_found(response)
            else:
                no_content(response)
        elif action is Action.accept:
            try:
                self._registrar.confirm(self._token)
            except LookupError:
                not_found(response)
            except AlreadySubscribedError:
                conflict(response, 'Already subscribed')
            else:
                no_content(response)
        elif action is Action.discard:
            # At least see if the token is in the database.
            pendable = self._pendings.confirm(self._token, expunge=True)
            if pendable is None:
                not_found(response)
            else:
                no_content(response)
        else:
            assert action is Action.reject, action
            # Like discard but sends a rejection notice to the user.
            pendable = self._pendings.confirm(self._token, expunge=True)
            if pendable is None:
                not_found(response)
            else:
                no_content(response)
                reason = arguments.get('reason', _('[No reason given]'))
                send_rejection(
                    self._mlist, _('Subscription request'),
                    pendable['email'], reason)


class _SubscriptionRequestsFound(_ModerationBase, CollectionMixin):
    """An abstract class to return a sub-set of subscription requests.

    This is used for filtering subscription requests based on token_owner.
    """
    def __init__(self, sub_requests):
        super().__init__()
        self._requests = sub_requests

    def _get_collection(self, requests):
        return self._requests


class _SubscriptionRequestCount:

    def __init__(self, mlist):
        self._mlist = mlist

    def on_get(self, request, response):
        validator = Validator(
            token_owner=enum_validator(TokenOwner),
            request_type=enum_validator(PendType),
            _optional=['token_owner', 'request_type'])
        try:
            data = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
        else:
            token_owner = data.pop('token_owner', None)
            pend_type = data.pop('request_type', PendType.subscription)
            count = getUtility(IPendings).count(
                mlist=self._mlist,
                pend_type=pend_type.name,
                token_owner=token_owner)
            okay(response, etag(dict(count=count)))


@public
class SubscriptionRequests(_ModerationBase, CollectionMixin):
    """Resource for membership change requests."""

    def __init__(self, mlist):
        super().__init__()
        self._mlist = mlist

    def on_get(self, request, response):
        """/lists/listname/requests"""
        validator = Validator(
            token_owner=enum_validator(TokenOwner),
            request_type=enum_validator(PendType),
            page=int,
            count=int,
            _optional=['token_owner', 'page', 'count', 'request_type'],
            )

        try:
            data = validator(request)
        except ValueError as error:
            bad_request(response, str(error))
        else:
            data.pop('page', None)
            data.pop('count', None)
            token_owner = data.pop('token_owner', None)
            pend_type = data.pop('request_type', PendType.subscription)
            pendings = getUtility(IPendings).find(
                mlist=self._mlist,
                pend_type=pend_type.name,
                token_owner=token_owner)
            resource = _SubscriptionRequestsFound(
                [token for token, pendable in pendings])
            okay(response, etag(resource._make_collection(request)))

    @child()
    def count(self, context, segments):
        return _SubscriptionRequestCount(self._mlist)

    @child(r'^(?P<token>[^/]+)')
    def subscription(self, context, segments, **kw):
        return IndividualRequest(self._mlist, kw['token'])



# Copyright (C) 2010-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Basic WSGI Application object for REST server."""

import re
import hmac
import logging

from base64 import b64decode
from falcon import App, HTTPUnauthorized
from falcon.routing import map_http_methods, set_default_responders
from mailman.config import config
from mailman.database.transaction import api_transaction
from mailman.rest.helpers import bad_request
from mailman.rest.root import Root
from public import public


log = logging.getLogger('mailman.http')

MISSING = object()
SLASH = '/'
EMPTYSTRING = ''
REALM = 'mailman3-rest'
UTF8 = 'utf-8'
WILDCARD_ACCEPT_HEADER = '*/*'


class Middleware:
    """Falcon middleware object for Mailman's REST API.

    This does two things.  It sets the API version on the resource
    object, and it verifies that the proper authentication has been
    performed.
    """
    def process_resource(self, request, response, resource, params):
        # Check the authorization credentials.
        authorized = False
        if request.auth is not None and request.auth.startswith('Basic '):
            # b64decode() returns bytes, but we require a str.
            credentials = b64decode(request.auth[6:]).decode('utf-8')
            username, password = credentials.split(':', 1)
            if (username == config.webservice.admin_user and
                    hmac.compare_digest(
                        password, config.webservice.admin_pass)):
                authorized = True
        if not authorized:
            # Not authorized.
            realm = 'Basic realm="{}",charset="{}"'.format(REALM, UTF8)
            raise HTTPUnauthorized(
                '401 Unauthorized',
                'REST API authorization failed',
                challenges=[realm])


def handle_ValueError(exc, request, response, params):
    """Handle ValueErrors in API code to return HTTPBadRequest.

    ValueErrors are raised often by Validator and should not return a 500 error
    resposne to the client.  This is a stop-gap for 500 errors due to
    ValueErrors, it is recommended that they be handled at the call-site,
    instead of here.
    """
    # Only handle ValueError, raise anything else right back.
    if not isinstance(exc, ValueError):
        raise exc
    bad_request(response, str(exc))


class ObjectRouter:
    def __init__(self, root):
        self._root = root

    def add_route(self, uri_template, method_map, resource):
        # We don't need this method for object-based routing.
        raise NotImplementedError

    def find(self, uri, req=None):
        segments = uri.split(SLASH)
        # Since the path is always rooted at /, skip the first segment, which
        # will always be the empty string.
        segments.pop(0)
        this_segment = segments.pop(0)
        resource = self._root
        context = {}
        while True:
            # Plumb the API through to all child resources.
            api = getattr(resource, 'api', None)
            # See if any of the resource's child links match the next segment.
            for name in dir(resource):
                if name.startswith('__') and name.endswith('__'):
                    continue
                attribute = getattr(resource, name, MISSING)
                assert attribute is not MISSING, name
                matcher = getattr(attribute, '__matcher__', MISSING)
                if matcher is MISSING:
                    continue
                result = None
                if isinstance(matcher, str):
                    # Is the matcher string a regular expression or plain
                    # string?  If it starts with a caret, it's a regexp.
                    if matcher.startswith('^'):
                        cre = re.compile(matcher)
                        # Search against the entire remaining path.
                        tmp_segments = segments[:]
                        tmp_segments.insert(0, this_segment)
                        remaining_path = SLASH.join(tmp_segments)
                        mo = cre.match(remaining_path)
                        if mo:
                            result = attribute(
                                context, segments, **mo.groupdict())
                    elif matcher == this_segment:
                        result = attribute(context, segments)
                else:
                    # The matcher is a callable.  It returns None if it
                    # doesn't match, and if it does, it returns a 3-tuple
                    # containing the positional arguments, the keyword
                    # arguments, and the remaining segments.  The attribute is
                    # then called with these arguments.  Note that the matcher
                    # wants to see the full remaining path components, which
                    # includes the current hop.
                    tmp_segments = segments[:]
                    tmp_segments.insert(0, this_segment)
                    matcher_result = matcher(tmp_segments)
                    if matcher_result is not None:
                        positional, keyword, segments = matcher_result
                        result = attribute(
                            context, segments, *positional, **keyword)
                # The attribute could return a 2-tuple giving the resource and
                # remaining path segments, or it could just return the result.
                # Of course, if the result is None, then the matcher did not
                # match.
                if result is None:
                    continue
                elif isinstance(result, tuple):
                    resource, segments = result
                else:
                    resource = result
                # See if the context set an API and set it on the next
                # resource in the chain, falling back to the parent resource's
                # API if there is one.
                resource.api = context.pop('api', api)
                # The method could have truncated the remaining segments,
                # meaning, it's consumed all the path segments, or this is the
                # last path segment.  In that case the resource we're left at
                # is the responder.
                if len(segments) == 0:
                    # We're at the end of the path, so the root must be the
                    # responder.
                    method_map = map_http_methods(resource)
                    set_default_responders(method_map)
                    return resource, method_map, context
                this_segment = segments.pop(0)
                break
            else:
                # None of the attributes matched this path component, so the
                # response is a 404.
                return None, None, None


class RootedAPI(App):
    def __init__(self, root, *args, **kws):
        super().__init__(
            *args,
            middleware=Middleware(),
            router=ObjectRouter(root),
            **kws)
        # Let Falcon parse the form data into the request object's
        # .params attribute.
        self.req_options.auto_parse_form_urlencoded = True
        # Don't ignore empty query parameters, e.g. preserve empty string
        # values, which some resources will interpret as a DELETE.
        self.req_options.keep_blank_qs_values = True

    # Override the base class implementation to wrap a transactional
    # handler around the call, so that the current transaction is
    # committed if no errors occur, and aborted otherwise.
    @api_transaction
    def __call__(self, environ, start_response):
        return super().__call__(environ, start_response)


@public
def make_application():
    """Return a callable WSGI application object."""
    app = RootedAPI(Root())
    app.add_error_handler(ValueError, handle_ValueError)
    return app



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST list tests."""

import unittest

from datetime import timedelta
from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.database.transaction import transaction
from mailman.interfaces.digests import DigestFrequency
from mailman.interfaces.listmanager import IListManager
from mailman.interfaces.mailinglist import IAcceptableAliasSet
from mailman.interfaces.member import DeliveryMode, MemberRole
from mailman.interfaces.template import ITemplateManager
from mailman.interfaces.usermanager import IUserManager
from mailman.model.mailinglist import AcceptableAlias
from mailman.runners.digest import DigestRunner
from mailman.testing.helpers import (
    call_api,
    get_queue_messages,
    make_testable_runner,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import RESTLayer
from mailman.utilities.datetime import now as right_now
from urllib.error import HTTPError
from zope.component import getUtility


class TestListsMissing(unittest.TestCase):
    """Test expected failures."""

    layer = RESTLayer

    def test_missing_list_roster_member_404(self):
        # /lists/<missing>/roster/member gives 404
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/missing@example.com'
                     '/roster/member')
        self.assertEqual(cm.exception.code, 404)

    def test_missing_list_roster_owner_404(self):
        # /lists/<missing>/roster/owner gives 404
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/missing@example.com'
                     '/roster/owner')
        self.assertEqual(cm.exception.code, 404)

    def test_missing_list_roster_moderator_404(self):
        # /lists/<missing>/roster/member gives 404
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/missing@example.com'
                     '/roster/moderator')
        self.assertEqual(cm.exception.code, 404)

    def test_missing_list_configuration_404(self):
        # /lists/<missing>/config gives 404
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/missing@example.com/config')
        self.assertEqual(cm.exception.code, 404)


class TestFindLists(unittest.TestCase):
    """Test /list/find"""

    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('test@example.com')
        self._user_manager = getUtility(IUserManager)
        self.address = self._user_manager.create_address('testing@example.com')

    def test_find_lists(self):
        # /lists/find gives 400 if no subscriber.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/lists/find')
        self.assertEqual(cm.exception.code, 400)

    def test_find_lists_with_subscriber(self):
        # Test GET /lists/find with a valid
        with transaction():
            self._mlist.subscribe(self.address)
            # We create another list so we have 2 lists.
            create_list('test2@example.com')
        json, response = call_api('http://localhost:9001/3.1/lists/find',
                                  {'subscriber': 'testing@example.com'})
        # This should return only one List.
        self.assertEqual(json['total_size'], 1)
        mlist = json['entries'][0]
        self.assertEqual(mlist['list_id'], 'test.example.com')

    def test_find_lists_with_subscriber_as_owner(self):
        # Test GET /lists/find with a valid subscriber and a valid role.
        with transaction():
            list2 = create_list('test2@example.com')
            create_list('test3@example.com')
            anne_addr = self._user_manager.create_address('anne@example.com')
            self._mlist.subscribe(anne_addr, role=MemberRole.owner)
            list2.subscribe(anne_addr, role=MemberRole.moderator)
        # With role=owner, only one list will be returned.
        json, response = call_api(
            'http://localhost:9001/3.1/lists/find',
            {'subscriber': 'anne@example.com', 'role': 'owner'})
        self.assertEqual(json['total_size'], 1)
        mlist = json['entries'][0]
        self.assertEqual(mlist['list_id'], 'test.example.com')
        # With role=moderator, we should get 2nd list.
        json, response = call_api(
            'http://localhost:9001/3.1/lists/find',
            {'subscriber': 'anne@example.com', 'role': 'moderator'})
        self.assertEqual(json['total_size'], 1)
        mlist = json['entries'][0]
        self.assertEqual(mlist['list_id'], 'test2.example.com')

    def test_find_lists_with_bad_role(self):
        with transaction():
            self._mlist.subscribe(self.address)
        # with role=badrole, we should get a 400 error with right reason set.
        with self.assertRaises(HTTPError) as cm:
            json, response = call_api(
                'http://localhost:9001/3.1/lists/find',
                {'subscriber': 'testing@example.com', 'role': 'badrole'})
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(
            cm.exception.reason,
            'Invalid Parameter "role": Accepted Values are:'
            ' member, owner, moderator, nonmember.')

    def test_find_lists_no_results(self):
        # No matching results exist for this query.
        with self.assertRaises(HTTPError) as cm:
            json, response = call_api(
                'http://localhost:9001/3.1/lists/find',
                {'subscriber': 'testing@example.com', 'role': 'owner'})
        self.assertEqual(cm.exception.code, 404)
        self.assertEqual(cm.exception.reason, '404 Not Found')

    def test_find_lists_bad_subscriber(self):
        # Bad subscriber should result in appropriate error.
        with self.assertRaises(HTTPError) as cm:
            json, response = call_api(
                'http://localhost:9001/3.1/lists/find',
                {'subscriber': 'testing.example.com', 'role': 'owner'})
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'Invalid Parameter "subscriber": .')


class TestLists(unittest.TestCase):
    """Test various aspects of mailing list resources."""

    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('test@example.com')
        self._usermanager = getUtility(IUserManager)

    def test_member_count_with_no_members(self):
        # The list initially has 0 members.
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test@example.com')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json['member_count'], 0)

    def test_member_count_with_one_member(self):
        # Add a member to a list and check that the resource reflects this.
        with transaction():
            anne = self._usermanager.create_address('anne@example.com')
            self._mlist.subscribe(anne)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test@example.com')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json['member_count'], 1)

    def test_member_count_with_two_members(self):
        # Add two members to a list and check that the resource reflects this.
        with transaction():
            anne = self._usermanager.create_address('anne@example.com')
            self._mlist.subscribe(anne)
            bart = self._usermanager.create_address('bar@example.com')
            self._mlist.subscribe(bart)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test@example.com')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json['member_count'], 2)

    def test_query_for_lists_in_missing_domain(self):
        # You cannot ask all the mailing lists in a non-existent domain.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/domains/no.example.org/lists')
        self.assertEqual(cm.exception.code, 404)

    def test_cannot_create_list_in_missing_domain(self):
        # You cannot create a mailing list in a domain that does not exist.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists', {
                     'fqdn_listname': 'ant@no-domain.example.org',
                     })
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'Domain does not exist: no-domain.example.org')

    def test_cannot_create_list_with_invalid_posting_address(self):
        # You cannot create a mailing list which would have an invalid list
        # posting address.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists', {
                     'fqdn_listname': '@example.com',
                     })
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'Invalid list posting address: @example.com')

    def test_create_list_failed_validator_returns_bad_requests(self):
        # Test that missing parameter or any other error that fails validation
        # doesn't raise 500 error.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists', method='POST')
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'Missing Parameter: fqdn_listname')

    def test_cannot_create_list_with_invalid_name(self):
        # You cannot create a mailing list which would have an invalid list
        # posting address.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists', {
                     'fqdn_listname': 'a/list@example.com',
                     })
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'Invalid list name: a/list')

    def test_cannot_create_duplicate_list(self):
        # You cannot create a list that already exists.
        call_api('http://localhost:9001/3.0/lists', {
                 'fqdn_listname': 'ant@example.com',
                 })
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists', {
                     'fqdn_listname': 'ant@example.com',
                     })
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason, 'Mailing list exists')

    def test_cannot_delete_missing_list(self):
        # You cannot delete a list that does not exist.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/bee.example.com',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)

    def test_cannot_delete_already_deleted_list(self):
        # You cannot delete a list twice.
        call_api('http://localhost:9001/3.0/lists', {
                 'fqdn_listname': 'ant@example.com',
                 })
        call_api('http://localhost:9001/3.0/lists/ant.example.com',
                 method='DELETE')
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com',
                     method='DELETE')
        self.assertEqual(cm.exception.code, 404)

    def test_roster(self):
        # Lists have rosters which can be accessed by role.
        with transaction():
            anne = self._usermanager.create_address('anne@example.com')
            bart = self._usermanager.create_address('bart@example.com')
            self._mlist.subscribe(anne)
            self._mlist.subscribe(bart)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test@example.com/roster/member')
        self.assertEqual(json['start'], 0)
        self.assertEqual(json['total_size'], 2)
        member = json['entries'][0]
        self.assertEqual(member['email'], 'anne@example.com')
        self.assertEqual(member['role'], 'member')
        member = json['entries'][1]
        self.assertEqual(member['email'], 'bart@example.com')
        self.assertEqual(member['role'], 'member')

    def test_delete_list_with_acceptable_aliases(self):
        # LP: #1432239 - deleting a mailing list with acceptable aliases
        # causes a SQLAlchemy error.  The aliases must be deleted first.
        with transaction():
            alias_set = IAcceptableAliasSet(self._mlist)
            alias_set.add('bee@example.com')
        call_api('http://localhost:9001/3.0/lists/test.example.com',
                 method='DELETE')
        # Neither the mailing list, nor the aliases are present.
        self.assertIsNone(getUtility(IListManager).get('test@example.com'))
        self.assertEqual(config.db.store.query(AcceptableAlias).count(), 0)

    def test_bad_roster_matcher(self):
        # Try to get a list's roster, but the roster name is bogus.
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com'
                     '/roster/bogus')
        self.assertEqual(cm.exception.code, 404)

    def test_bad_config_matcher(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com'
                     '/config/volume/bogus')
        self.assertEqual(cm.exception.code, 404)

    def test_bad_list_get(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/bogus.example.com')
        self.assertEqual(cm.exception.code, 404)

    def test_not_found_member_role(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/test.example.com'
                     '/owner/nobody@example.com')
        self.assertEqual(cm.exception.code, 404)

    def test_list_mass_unsubscribe_all_succeed(self):
        with transaction():
            aperson = self._usermanager.create_address('aperson@example.com')
            bperson = self._usermanager.create_address('bperson@example.com')
            cperson = self._usermanager.create_address('cperson@example.com')
            self._mlist.subscribe(aperson)
            self._mlist.subscribe(bperson)
            self._mlist.subscribe(cperson)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test.example.com'
            '/roster/member', {
                'emails': ['aperson@example.com',
                           'bperson@example.com',
                           ]},
            'DELETE')
        self.assertEqual(response.status_code, 200)
        # Remove variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'aperson@example.com': True,
            'bperson@example.com': True,
            })

    def test_list_mass_unsubscribe_all_fail(self):
        with transaction():
            aperson = self._usermanager.create_address('aperson@example.com')
            bperson = self._usermanager.create_address('bperson@example.com')
            cperson = self._usermanager.create_address('cperson@example.com')
            self._mlist.subscribe(aperson)
            self._mlist.subscribe(bperson)
            self._mlist.subscribe(cperson)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test.example.com'
            '/roster/member', {
                'emails': ['yperson@example.com',
                           'zperson@example.com',
                           ]},
            'DELETE')
        self.assertEqual(response.status_code, 200)
        # Remove variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'yperson@example.com': False,
            'zperson@example.com': False,
            })

    def test_list_mass_unsubscribe_mixed_success(self):
        with transaction():
            aperson = self._usermanager.create_address('aperson@example.com')
            bperson = self._usermanager.create_address('bperson@example.com')
            cperson = self._usermanager.create_address('cperson@example.com')
            self._mlist.subscribe(aperson)
            self._mlist.subscribe(bperson)
            self._mlist.subscribe(cperson)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test.example.com'
            '/roster/member', {
                'emails': ['aperson@example.com',
                           'zperson@example.com',
                           ]},
            'DELETE')
        self.assertEqual(response.status_code, 200)
        # Remove variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'aperson@example.com': True,
            'zperson@example.com': False,
            })

    def test_list_mass_unsubscribe_with_duplicates(self):
        with transaction():
            aperson = self._usermanager.create_address('aperson@example.com')
            bperson = self._usermanager.create_address('bperson@example.com')
            cperson = self._usermanager.create_address('cperson@example.com')
            self._mlist.subscribe(aperson)
            self._mlist.subscribe(bperson)
            self._mlist.subscribe(cperson)
        json, response = call_api(
            'http://localhost:9001/3.0/lists/test.example.com'
            '/roster/member', {
                'emails': ['aperson@example.com',
                           'aperson@example.com',
                           'bperson@example.com',
                           'zperson@example.com',
                           ]},
            'DELETE')
        self.assertEqual(response.status_code, 200)
        # Remove variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'aperson@example.com': True,
            'bperson@example.com': True,
            'zperson@example.com': False,
            })

    def test_list_mass_unsubscribe_bogus_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/bogus.example.com'
                     '/roster/member',
                     None, 'DELETE')
        self.assertEqual(cm.exception.code, 404)

    def test_list_mass_unsubscribe_with_no_data(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/test.example.com'
                     '/roster/member',
                     None, 'DELETE')
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual('Missing Parameter: emails', cm.exception.reason)


class TestListStyles(unittest.TestCase):
    """Test /lists/styles."""

    layer = RESTLayer

    def test_styles(self):
        json, response = call_api('http://localhost:9001/3.0/lists/styles')
        self.assertEqual(response.status_code, 200)
        # Remove the variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'style_names': [
                'legacy-announce', 'legacy-default', 'private-default'
                ],
            'styles': [
                {'name': 'legacy-announce',
                 'description': 'Announce only mailing list style.'},
                {'name': 'legacy-default',
                 'description': 'Ordinary discussion mailing list style.'},
                {'name': 'private-default',
                 'description': 'Discussion mailing list style with ' +
                 'private archives.'},
                    ],
            'default': 'legacy-default'
            })


class TestListArchivers(unittest.TestCase):
    """Test corner cases for list archivers."""

    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('ant@example.com')

    def test_archiver_statuses(self):
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/archivers')
        self.assertEqual(response.status_code, 200)
        # Remove the variable data.
        json.pop('http_etag')
        self.assertEqual(json, {
            'mail-archive': True,
            'mhonarc': True,
            })

    def test_archiver_statuses_on_missing_lists(self):
        # You cannot get the archiver statuses on a list that doesn't exist.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/bee.example.com/archivers')
        self.assertEqual(cm.exception.code, 404)

    def test_put_bogus_archiver(self):
        # You cannot PUT on an archiver the list doesn't know about.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/archivers', {
                    'bogus-archiver': True,
                    },
                method='PUT')
        self.assertEqual(cm.exception.code, 400)
        self.assertTrue(
            'Unexpected parameters: bogus-archiver' in cm.exception.reason,)

    def test_patch_bogus_archiver(self):
        # You cannot PATCH on an archiver the list doesn't know about.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/archivers', {
                    'bogus-archiver': True,
                    },
                method='PATCH')
        self.assertEqual(cm.exception.code, 400)
        self.assertTrue(
            'Unexpected parameters: bogus-archiver' in cm.exception.reason)

    def test_put_incomplete_statuses(self):
        # PUT requires the full resource representation.  This one forgets to
        # specify the mhonarc archiver.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/archivers', {
                    'mail-archive': True,
                    },
                method='PUT')
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual('Missing Parameter: mhonarc', cm.exception.reason,)

    def test_patch_bogus_status(self):
        # Archiver statuses must be interpretable as booleans.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/archivers', {
                    'mail-archive': 'sure',
                    'mhonarc': 'no'
                    },
                method='PATCH')
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual('Invalid boolean value: sure', cm.exception.reason)


class TestListPagination(unittest.TestCase):
    """Test mailing list pagination functionality.

    We create a bunch of mailing lists within a domain.  When we want to
    get all the lists in that domain via the REST API, we need to
    paginate over them, otherwise there could be too many for display.
    """

    layer = RESTLayer

    def setUp(self):
        with transaction():
            # Create a bunch of mailing lists in the example.com domain.
            create_list('ant@example.com')
            create_list('bee@example.com')
            create_list('cat@example.com')
            create_list('dog@example.com')
            create_list('emu@example.com')
            create_list('fly@example.com')

    def test_first_page(self):
        json, response = call_api(
            'http://localhost:9001/3.0/domains/example.com/lists'
            '?count=1&page=1')
        # There are 6 total lists, but only the first one in the page.
        self.assertEqual(json['total_size'], 6)
        self.assertEqual(json['start'], 0)
        self.assertEqual(len(json['entries']), 1)
        entry = json['entries'][0]
        self.assertEqual(entry['fqdn_listname'], 'ant@example.com')

    def test_second_page(self):
        json, response = call_api(
            'http://localhost:9001/3.0/domains/example.com/lists'
            '?count=1&page=2')
        # There are 6 total lists, but only the first one in the page.
        self.assertEqual(json['total_size'], 6)
        self.assertEqual(json['start'], 1)
        self.assertEqual(len(json['entries']), 1)
        entry = json['entries'][0]
        self.assertEqual(entry['fqdn_listname'], 'bee@example.com')

    def test_last_page(self):
        json, response = call_api(
            'http://localhost:9001/3.0/domains/example.com/lists'
            '?count=1&page=6')
        # There are 6 total lists, but only the first one in the page.
        self.assertEqual(json['total_size'], 6)
        self.assertEqual(json['start'], 5)
        self.assertEqual(len(json['entries']), 1)
        entry = json['entries'][0]
        self.assertEqual(entry['fqdn_listname'], 'fly@example.com')

    def test_zeroth_page(self):
        # Page numbers start at one.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/domains/example.com/lists'
                '?count=1&page=0')
        self.assertEqual(cm.exception.code, 400)

    def test_negative_page(self):
        # Negative pages are not allowed.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/domains/example.com/lists'
                '?count=1&page=-1')
        self.assertEqual(cm.exception.code, 400)

    def test_past_last_page(self):
        # The 7th page doesn't exist so the collection is empty.
        json, response = call_api(
            'http://localhost:9001/3.0/domains/example.com/lists'
            '?count=1&page=7')
        # There are 6 total lists, but only the first one in the page.
        self.assertEqual(json['total_size'], 6)
        self.assertEqual(json['start'], 6)
        self.assertNotIn('entries', json)


class TestListDigests(unittest.TestCase):
    """Test /lists/<list-id>/digest"""

    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('ant@example.com')
            self._mlist.send_goodbye_message = False
            self._mlist.send_welcome_message = False
            anne = getUtility(IUserManager).create_address('anne@example.com')
            self._mlist.subscribe(anne)
            anne.preferences.delivery_mode = DeliveryMode.plaintext_digests

    def test_bad_digest_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/bogus.example.com/digest')
        self.assertEqual(cm.exception.code, 404)

    def test_post_nothing_to_do(self):
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/digest', {})
        self.assertEqual(response.status_code, 200)

    def test_post_something_to_do(self):
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                bump=True))
        self.assertEqual(response.status_code, 202)

    def test_post_bad_request(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                    bogus=True))
        self.assertEqual(cm.exception.code, 400)
        self.assertTrue('Unexpected parameters: bogus' in cm.exception.reason)

    def test_post_both_send_and_periodic_options(self):
        # send and periodic options cannot both be set at the same time.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                    send=True,
                    periodic=True))
        self.assertEqual(cm.exception.code, 400)
        self.assertEqual(cm.exception.reason,
                         'send and periodic options are mutually exclusive')

    def test_bump_before_send(self):
        with transaction():
            self._mlist.digest_volume_frequency = DigestFrequency.monthly
            self._mlist.volume = 7
            self._mlist.next_digest_number = 4
            self._mlist.digest_last_sent_at = right_now() + timedelta(
                days=-32)
        msg = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: message 1

""")
        config.handlers['to-digest'].process(self._mlist, msg, {})
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                send=True,
                bump=True))
        self.assertEqual(response.status_code, 202)
        make_testable_runner(DigestRunner, 'digest').run()
        # The volume is 8 and the digest number is 2 because a digest was sent
        # after the volume/number was bumped.
        self.assertEqual(self._mlist.volume, 8)
        self.assertEqual(self._mlist.next_digest_number, 2)
        self.assertEqual(self._mlist.digest_last_sent_at, right_now())
        items = get_queue_messages('virgin')
        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].msg['subject'], 'Ant Digest, Vol 8, Issue 1')

    def test_periodic_digests_send(self):
        with transaction():
            self._mlist.digest_send_periodic = False
        msg = mfs("""\
To: ant@example.com
From: anne@example.com
Subject: message 2

""")
        config.handlers['to-digest'].process(self._mlist, msg, {})
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                periodic=True))
        # There shouldn't be any message going out after the above call.
        self.assertEqual(response.status_code, 202)
        make_testable_runner(DigestRunner, 'digest').run()
        items = get_queue_messages('virgin')
        self.assertEqual(len(items), 0)
        # Now, we set the periodic_send_digest to True and call the API again.
        with transaction():
            self._mlist.digest_send_periodic = True
        json, response = call_api(
            'http://localhost:9001/3.0/lists/ant.example.com/digest', dict(
                periodic=True))
        # Now, there should be one message in the virgin queue to be sent out.
        make_testable_runner(DigestRunner, 'digest').run()
        items = get_queue_messages('virgin')
        self.assertEqual(len(items), 1)
        # This should be the first volume of the digest going out.
        self.assertEqual(str(items[0].msg['subject']),
                         'Ant Digest, Vol 1, Issue 1')


class TestListTemplates(unittest.TestCase):
    """Test /lists/<list-id>/uris"""

    layer = RESTLayer

    def setUp(self):
        with transaction():
            self._mlist = create_list('ant@example.com')

    def test_no_templates_for_api_30(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.0/lists/ant.example.com/uris')
        self.assertEqual(cm.exception.code, 404)

    def test_no_templates_for_missing_list(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/lists/bee.example.com/uris')
        self.assertEqual(cm.exception.code, 404)

    def test_path_too_long(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/lists/ant.example.com/uris'
                     '/foo/bar')
        self.assertEqual(cm.exception.code, 400)

    def test_get_unknown_uri(self):
        with self.assertRaises(HTTPError) as cm:
            call_api('http://localhost:9001/3.1/lists/ant.example.com/uris'
                     '/not:a:template')
        self.assertEqual(cm.exception.code, 404)

    def test_get_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', 'ant.example.com',
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', 'ant.example.com',
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json['start'], 0)
        self.assertEqual(json['total_size'], 2)
        self.assertEqual(
            json['self_link'],
            'http://localhost:9001/3.1/lists/ant.example.com/uris')
        self.assertEqual(json['entries'], [
            {'http_etag': '"6612187ed6604ce54a57405fd66742557391ed4a"',
             'name': 'list:user:notice:goodbye',
             'password': 'the password',
             'self_link': ('http://localhost:9001/3.1/lists/ant.example.com'
                           '/uris/list:user:notice:goodbye'),
             'uri': 'http://example.com/goodbye',
             'username': 'a user',
             },
            {'http_etag': '"cb1ab5eee2242143d2984edd0487532915ad3a8e"',
             'name': 'list:user:notice:welcome',
             'self_link': ('http://localhost:9001/3.1/lists/ant.example.com'
                           '/uris/list:user:notice:welcome'),
             'uri': 'http://example.com/welcome',
             }])

    def test_patch_uris(self):
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')
        template = manager.raw('list:user:notice:goodbye', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertIsNone(template.username)
        self.assertEqual(template.password, '')

    def test_patch_uris_with_credentials(self):
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris', {
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PATCH')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:user:notice:welcome', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/welcome')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_patch_uris_with_partial_credentials(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/lists/ant.example.com/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'username': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_put_all_uris(self):
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris', {
                'list:admin:action:post': '',
                'list:admin:action:subscribe': '',
                'list:admin:action:unsubscribe': '',
                'list:admin:notice:subscribe': '',
                'list:admin:notice:unrecognized': '',
                'list:admin:notice:unsubscribe': '',
                'list:admin:notice:pending': '',
                'list:admin:notice:increment': '',
                'list:admin:notice:disable': '',
                'list:admin:notice:removal': '',
                'list:member:digest:footer': '',
                'list:member:digest:header': '',
                'list:member:digest:masthead': '',
                'list:member:regular:footer': 'http://example.org/footer',
                'list:member:regular:header': 'http://example.org/header',
                'list:user:action:invite': '',
                'list:user:action:subscribe': '',
                'list:user:action:unsubscribe': '',
                'list:user:notice:goodbye': 'http://example.org/goodbye',
                'list:user:notice:hold': '',
                'list:user:notice:no-more-today': '',
                'list:user:notice:post': '',
                'list:user:notice:probe': '',
                'list:user:notice:refuse': '',
                'list:user:notice:rejected': '',
                'list:user:notice:welcome': 'http://example.org/welcome',
                'list:user:notice:warning': '',
                'password': 'some password',
                'username': 'anne.person',
                }, method='PUT')
        self.assertEqual(response.status_code, 204)
        manager = getUtility(ITemplateManager)
        template = manager.raw('list:member:digest:footer', 'ant.example.com')
        self.assertIsNone(template)
        template = manager.raw('list:member:digest:header', 'ant.example.com')
        self.assertIsNone(template)
        template = manager.raw('list:member:regular:footer', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/footer')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:member:regular:header', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/header')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')
        template = manager.raw('list:user:notice:goodbye', 'ant.example.com')
        self.assertEqual(template.uri, 'http://example.org/goodbye')
        self.assertEqual(template.username, 'anne.person')
        self.assertEqual(template.password, 'some password')

    def test_delete_all_uris(self):
        manager = getUtility(ITemplateManager)
        with transaction():
            manager.set(
                'list:user:notice:welcome', 'ant.example.com',
                'http://example.com/welcome')
            manager.set(
                'list:user:notice:goodbye', 'ant.example.com',
                'http://example.com/goodbye',
                'a user', 'the password',
                )
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris',
            method='DELETE')
        self.assertEqual(response.status_code, 204)
        self.assertIsNone(
            manager.raw('list:user:notice:welcome', 'ant.example.com'))
        self.assertIsNone(
            manager.raw('list:user:notice:goodbye', 'ant.example.com'))

    def test_get_a_url(self):
        with transaction():
            getUtility(ITemplateManager).set(
                'list:user:notice:welcome', 'ant.example.com',
                'http://example.com/welcome')
        json, response = call_api(
            'http://localhost:9001/3.1/lists/ant.example.com/uris'
            '/list:user:notice:welcome')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json, {
            'http_etag': '"36f8bef800cfd278f097c61c5892a34c0650f4aa"',
            'self_link': ('http://localhost:9001/3.1/lists/ant.example.com'
                          '/uris/list:user:notice:welcome'),
            'uri': 'http://example.com/welcome',
            })

    def test_get_a_bad_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/lists/ant.example.com/uris'
                '/list:user:notice:notemplate')
        self.assertEqual(cm.exception.code, 404)

    def test_get_unset_url(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/lists/ant.example.com/uris'
                '/list:user:notice:welcome')
        self.assertEqual(cm.exception.code, 404)

    def test_patch_url_with_too_many_parameters(self):
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.1/lists/ant.example.com/uris', {
                    'list:user:notice:welcome': 'http://example.org/welcome',
                    'list:user:notice:goodbye': 'http://example.org/goodbye',
                    'secret': 'some password',
                    'person': 'anne.person',
                    }, method='PATCH')
        self.assertEqual(cm.exception.code, 400)

    def test_deprecated_resources(self):
        # This resource does not exist with API 3.0.
        with self.assertRaises(HTTPError) as cm:
            call_api(
                'http://localhost:9001/3.0/templates/ant@example.com'
                '/footer/en')
        self.assertEqual(cm.exception.code, 404)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the withlist/shell command."""

import os
import unittest

from click.testing import CliRunner
from contextlib import ExitStack
from mailman.app.lifecycle import create_list
from mailman.commands.cli_withlist import shell
from mailman.config import config
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import configuration
from mailman.testing.layers import ConfigLayer
from mailman.utilities.modules import hacked_sys_modules
from types import ModuleType
from unittest.mock import MagicMock, patch


try:
    import readline  # noqa: F401
    has_readline = True
except ImportError:
    has_readline = False


class TestShell(unittest.TestCase):
    layer = ConfigLayer
    maxDiff = None

    def setUp(self):
        self._command = CliRunner()

    def test_namespace(self):
        with patch('mailman.commands.cli_withlist.start_python') as mock:
            self._command.invoke(shell, ('--interactive',))
            self.assertEqual(mock.call_count, 1)
            # Don't test that all names are available, just a few choice ones.
            positional, keywords = mock.call_args
            namespace = positional[0]
            self.assertIn('getUtility', namespace)
            self.assertIn('IArchiver', namespace)
            self.assertEqual(namespace['IUserManager'], IUserManager)

    @configuration('shell', banner='my banner')
    def test_banner(self):
        with patch('mailman.commands.cli_withlist.interact') as mock:
            self._command.invoke(shell, ('--interactive',))
            self.assertEqual(mock.call_count, 1)
            positional, keywords = mock.call_args
            self.assertEqual(keywords['banner'], 'my banner\n')

    @unittest.skipUnless(has_readline, 'readline module is not available')
    @configuration('shell', history_file='$var_dir/history.py')
    def test_history_file(self):
        with patch('mailman.commands.cli_withlist.interact'):
            self._command.invoke(shell, ('--interactive',))
            history_file = os.path.join(config.VAR_DIR, 'history.py')
            self.assertTrue(os.path.exists(history_file))

    @configuration('shell', use_ipython='yes')
    def test_start_ipython4(self):
        mock = MagicMock()
        with hacked_sys_modules('IPython.terminal.embed', mock):
            self._command.invoke(shell, ('--interactive',))
        posargs, kws = mock.InteractiveShellEmbed.instance().mainloop.call_args
        self.assertEqual(
            kws['display_banner'], """Welcome to the GNU Mailman shell
Use commit() to commit changes.
Use abort() to discard changes since the last commit.
Exit with ctrl+D does an implicit commit() but exit() does not.\n""")

    @configuration('shell', use_ipython='yes')
    def test_start_ipython1(self):
        mock = MagicMock()
        with hacked_sys_modules('IPython.frontend.terminal.embed', mock):
            self._command.invoke(shell, ('--interactive',))
        posargs, kws = mock.InteractiveShellEmbed.instance.call_args
        self.assertEqual(
            kws['banner1'], """Welcome to the GNU Mailman shell
Use commit() to commit changes.
Use abort() to discard changes since the last commit.
Exit with ctrl+D does an implicit commit() but exit() does not.\n""")

    @configuration('shell', use_ipython='debug')
    def test_start_ipython_debug(self):
        mock = MagicMock()
        with hacked_sys_modules('IPython.terminal.embed', mock):
            self._command.invoke(shell, ('--interactive',))
        posargs, kws = mock.InteractiveShellEmbed.instance().mainloop.call_args
        self.assertEqual(
            kws['display_banner'], """Welcome to the GNU Mailman shell
Use commit() to commit changes.
Use abort() to discard changes since the last commit.
Exit with ctrl+D does an implicit commit() but exit() does not.\n""")

    @configuration('shell', use_ipython='oops')
    def test_start_ipython_invalid(self):
        mock = MagicMock()
        with hacked_sys_modules('IPython.terminal.embed', mock):
            results = self._command.invoke(shell, ('--interactive',))
        self.assertEqual(
            results.output,
            'Invalid value for [shell]use_python: oops\n')
        # mainloop() never got called.
        self.assertIsNone(
            mock.InteractiveShellEmbed.instance().mainloop.call_args)

    @configuration('shell', use_ipython='yes')
    def test_start_ipython_uninstalled(self):
        with ExitStack() as resources:
            # Pretend iPython isn't available at all.
            resources.enter_context(patch(
                'mailman.commands.cli_withlist.start_ipython1',
                return_value=None))
            resources.enter_context(patch(
                'mailman.commands.cli_withlist.start_ipython4',
                return_value=None))
            results = self._command.invoke(shell, ('--interactive',))
        self.assertEqual(
            results.output,
            'ipython is not available, set use_ipython to no\n')

    def test_regex_without_run(self):
        results = self._command.invoke(shell, ('-l', '^.*example.com'))
        self.assertEqual(results.exit_code, 2)
        self.assertEqual(
            results.output,
            'Usage: shell [OPTIONS] [RUN_ARGS]...\n'
            'Try \'shell --help\' for help.\n\n'
            'Error: Regular expression requires --run\n')

    def test_listspec_without_run(self):
        create_list('ant@example.com')
        mock = MagicMock()
        with ExitStack() as resources:
            resources.enter_context(
                hacked_sys_modules('IPython.terminal.embed', mock))
            interactive_mock = resources.enter_context(patch(
                'mailman.commands.cli_withlist.do_interactive'))
            self._command.invoke(shell, ('-l', 'ant.example.com'))
        posargs, kws = interactive_mock.call_args
        self.assertEqual(
            posargs[1],
            'The variable \'m\' is the ant.example.com mailing list')

    def test_listspec_without_run_no_such_list(self):
        results = self._command.invoke(shell, ('-l', 'ant.example.com'))
        self.assertEqual(results.exit_code, 2)
        self.assertEqual(
            results.output,
            'Usage: shell [OPTIONS] [RUN_ARGS]...\n'
            'Try \'shell --help\' for help.\n\n'
            'Error: No such list: ant.example.com\n')

    def test_run_without_listspec(self):
        something = ModuleType('something')
        something.something = lambda: print('I am a something!')
        with hacked_sys_modules('something', something):
            results = self._command.invoke(shell, ('--run', 'something'))
        self.assertEqual(results.exit_code, 0)
        self.assertEqual(results.output, 'I am a something!\n')

    def test_run_bogus_listspec(self):
        results = self._command.invoke(
            shell, ('-l', 'bee.example.com', '--run', 'something'))
        self.assertEqual(results.exit_code, 2)
        self.assertEqual(
            results.output,
            'Usage: shell [OPTIONS] [RUN_ARGS]...\n'
            'Try \'shell --help\' for help.\n\n'
            'Error: No such list: bee.example.com\n')



# Copyright (C) 2012-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test styles."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.interfaces.archiver import ArchivePolicy
from mailman.interfaces.mailinglist import SubscriptionPolicy
from mailman.interfaces.styles import (
    DuplicateStyleError,
    IStyle,
    IStyleManager,
)
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility
from zope.interface import implementer, Invalid


@implementer(IStyle)
class DummyStyle:

    name = 'dummy'
    description = 'A dummy style.'

    def apply(self, mlist):
        pass


class TestStyle(unittest.TestCase):
    """Test styles."""

    layer = ConfigLayer

    def setUp(self):
        self.manager = getUtility(IStyleManager)

    def test_register_style_again(self):
        # Registering a style with the same name as a previous style raises an
        # exception.
        self.manager.register(DummyStyle())
        self.assertRaises(DuplicateStyleError,
                          self.manager.register, DummyStyle())

    def test_register_a_non_style(self):
        # You can't register something that doesn't implement the IStyle
        # interface.
        self.assertRaises(Invalid, self.manager.register, object())

    def test_unregister_a_non_registered_style(self):
        # You cannot unregister a style that hasn't yet been registered.
        self.assertRaises(KeyError,
                          self.manager.unregister, DummyStyle())


class TestPrivateDefaultStyle(unittest.TestCase):
    """Test PrivateDefaultStyle."""

    layer = ConfigLayer

    def setUp(self):
        self.mlist = create_list('test-list@example.com')
        self.manager = getUtility(IStyleManager)

    def test_private_default(self):
        self.manager.get('private-default').apply(self.mlist)
        self.assertEqual(self.mlist.advertised, False)
        self.assertEqual(self.mlist.archive_policy, ArchivePolicy.private)
        self.assertEqual(self.mlist.subscription_policy,
                         SubscriptionPolicy.confirm_then_moderate)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `suspicious` rule."""


import unittest

from email.header import Header
from mailman.app.lifecycle import create_list
from mailman.email.message import Message
from mailman.rules import suspicious
from mailman.testing.layers import ConfigLayer


class TestSuspicious(unittest.TestCase):
    """Test the suspicious rule."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._rule = suspicious.SuspiciousHeader()

    def test_header_instance(self):
        msg = Message()
        msg['From'] = Header('user@example.com')
        self._mlist.bounce_matching_headers = 'from: spam@example.com'
        result = self._rule.check(self._mlist, msg, {})
        self.assertFalse(result)

    def test_suspicious_returns_reason(self):
        msg = Message()
        msg['From'] = Header('spam@example.com')
        self._mlist.bounce_matching_headers = 'from: spam@example.com'
        msgdata = {}
        result = self._rule.check(self._mlist, msg, msgdata)
        self.assertTrue(result)
        self.assertEqual(
            msgdata['moderation_reasons'],
            [('Header "{}" matched a bounce_matching_header line',
              'spam@example.com')]
            )



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Creation/deletion hooks for the MTAs that do things automatically.

Exim is one example of an MTA that Just Works.
"""

from mailman.interfaces.mta import IMailTransportAgentLifecycle
from public import public
from zope.interface import implementer


@public
@implementer(IMailTransportAgentLifecycle)
class NullMTA:
    """Null MTA that just satisfies the interface."""

    def create(self, mlist):
        """See `IMailTransportAgentLifecycle`."""
        pass

    def delete(self, mlist):
        """See `IMailTransportAgentLifecycle`."""
        pass

    def regenerate(self, output=None):
        """See `IMailTransportAgentLifecycle`."""
        pass



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Testing app.inject functions."""

import unittest

from email import message_from_bytes
from mailman.app.inject import inject_message, inject_text
from mailman.app.lifecycle import create_list
from mailman.email.message import Message
from mailman.testing.helpers import (
    get_queue_messages,
    specialized_message_from_string as message_from_string,
)
from mailman.testing.layers import ConfigLayer


NL = '\n'


class TestInjectMessage(unittest.TestCase):
    """Test message injection."""

    layer = ConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com')
        self.msg = message_from_string("""\
From: anne@example.com
To: test@example.com
Subject: A test message
Message-ID: <first>
Date: Tue, 14 Jun 2011 21:12:00 -0400

Nothing.
""")
        self.msg2 = message_from_bytes("""
From: anne@example.com
To: test@example.com
Subject: A test message
Message-ID: <first>
Date: Tue, 14 Jun 2011 21:12:00 -0400
MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 8bit

Here's some text.
Throw in some specials “fancy quoted”
""".encode('utf-8'), Message)
        # Let assertMultiLineEqual work without bounds.

    def test_inject_message(self):
        # Test basic inject_message() call.
        inject_message(self.mlist, self.msg)
        items = get_queue_messages('in', expected_count=1)
        self.assertMultiLineEqual(items[0].msg.as_string(),
                                  self.msg.as_string())
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msgdata['original_size'],
                         len(self.msg.as_string()))

    def test_inject_message_with_recipients(self):
        # Explicit recipients end up in the metadata.
        recipients = ['bart@example.com', 'cris@example.com']
        inject_message(self.mlist, self.msg, recipients)
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msgdata['recipients'], recipients)

    def test_inject_message_to_queue(self):
        # Explicitly use a different queue.
        inject_message(self.mlist, self.msg, switchboard='virgin')
        get_queue_messages('in', expected_count=0)
        items = get_queue_messages('virgin', expected_count=1)
        self.assertMultiLineEqual(items[0].msg.as_string(),
                                  self.msg.as_string())
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msgdata['original_size'],
                         len(self.msg.as_string()))

    def test_inject_message_without_message_id(self):
        # inject_message() adds a Message-ID header if it's missing.
        del self.msg['message-id']
        self.assertNotIn('message-id', self.msg)
        inject_message(self.mlist, self.msg)
        self.assertIn('message-id', self.msg)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('message-id', items[0].msg)
        self.assertEqual(items[0].msg['message-id'], self.msg['message-id'])

    def test_inject_message_without_date(self):
        # inject_message() adds a Date header if it's missing.
        del self.msg['date']
        self.assertNotIn('date', self.msg)
        inject_message(self.mlist, self.msg)
        self.assertIn('date', self.msg)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('date', items[0].msg)
        self.assertEqual(items[0].msg['date'], self.msg['date'])

    def test_inject_message_with_keywords(self):
        # Keyword arguments are copied into the metadata.
        inject_message(self.mlist, self.msg, foo='yes', bar='no')
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msgdata['foo'], 'yes')
        self.assertEqual(items[0].msgdata['bar'], 'no')

    def test_inject_message_id_hash(self):
        # When the injected message has a Message-ID header, the injected
        # message will also get a Message-ID-Hash header.
        inject_message(self.mlist, self.msg)
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msg['message-id-hash'],
                         '4CMWUN6BHVCMHMDAOSJZ2Q72G5M32MWB')

    def test_inject_message_id_hash_without_message_id(self):
        # When the injected message does not have a Message-ID header, a
        # Message-ID header will be added, and the injected message will also
        # get a Message-ID-Hash header.
        del self.msg['message-id']
        self.assertNotIn('message-id', self.msg)
        self.assertNotIn('message-id-hash', self.msg)
        inject_message(self.mlist, self.msg)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('message-id', items[0].msg)
        self.assertIn('message-id-hash', items[0].msg)

    def test_inject_non_ascii_message(self):
        # Test basic inject_message() call with a non-ascii message.
        inject_message(self.mlist, self.msg2)
        items = get_queue_messages('in', expected_count=1)
        self.assertMultiLineEqual(items[0].msg.as_string(),
                                  self.msg2.as_string())
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msgdata['original_size'],
                         len(self.msg2.as_string()))


class TestInjectText(unittest.TestCase):
    """Test text injection."""

    layer = ConfigLayer
    maxDiff = None

    def setUp(self):
        self.mlist = create_list('test@example.com')
        self.text = """\
From: bart@example.com
To: test@example.com
Subject: A test message
Message-ID: <second>
Date: Tue, 14 Jun 2011 21:12:00 -0400

Nothing.
"""
        self.text2 = """\
From: bart@example.com
To: test@example.com
Subject: A test message
Message-ID: <second>
Date: Tue, 14 Jun 2011 21:12:00 -0400
MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: 8bit

Here's some text.
Throw in some specials “fancy quoted”
"""

    def _remove_line(self, header):
        return NL.join(line for line in self.text.splitlines()
                       if not line.lower().startswith(header))

    def test_inject_text(self):
        # Test basic inject_text() call.
        inject_text(self.mlist, self.text)
        items = get_queue_messages('in', expected_count=1)
        self.assertIsInstance(items[0].msg, Message)
        self.assertEqual(items[0].msg['message-id-hash'],
                         'GUXXQKNCHBFQAHGBFMGCME6HKZCUUH3K')
        # Delete these headers because they don't exist in the original text.
        del items[0].msg['message-id-hash']
        del items[0].msg['x-message-id-hash']
        self.assertMultiLineEqual(items[0].msg.as_string(), self.text)
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msgdata['original_size'],
                         # Add back the Message-ID-Hash and X-Message-ID-Hash
                         # headers which wer in the message contributing to the
                         # original_size, but weren't in the original text.
                         # Don't forget the space, delimeter, and newline!
                         len(self.text) + 50 + 52)

    def test_inject_text_with_recipients(self):
        # Explicit recipients end up in the metadata.
        recipients = ['bart@example.com', 'cris@example.com']
        inject_text(self.mlist, self.text, recipients)
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msgdata['recipients'], recipients)

    def test_inject_text_to_queue(self):
        # Explicitly use a different queue.
        inject_text(self.mlist, self.text, switchboard='virgin')
        get_queue_messages('in', expected_count=0)
        items = get_queue_messages('virgin', expected_count=1)
        # Remove the Message-ID-Hash header which isn't in the original text.
        del items[0].msg['message-id-hash']
        del items[0].msg['x-message-id-hash']
        self.assertMultiLineEqual(items[0].msg.as_string(), self.text)
        self.assertEqual(items[0].msgdata['listid'], 'test.example.com')
        self.assertEqual(items[0].msgdata['original_size'],
                         # Add back the Message-ID-Hash and X-Message-ID-Hash
                         # headers which wer in the message contributing to the
                         # original_size, but weren't in the original text.
                         # Don't forget the space, delimeter, and newline!
                         len(self.text) + 50 + 52)

    def test_inject_text_without_message_id(self):
        # inject_text() adds a Message-ID header if it's missing.
        filtered = self._remove_line('message-id')
        self.assertNotIn('Message-ID', filtered)
        inject_text(self.mlist, filtered)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('message-id', items[0].msg)

    def test_inject_text_without_date(self):
        # inject_text() adds a Date header if it's missing.
        filtered = self._remove_line('date')
        self.assertNotIn('date', filtered)
        inject_text(self.mlist, self.text)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('date', items[0].msg)

    def test_inject_text_adds_original_size(self):
        # The metadata gets an original_size attribute that is the length of
        # the injected text.
        inject_text(self.mlist, self.text)
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msgdata['original_size'],
                         # Add back the Message-ID-Hash and X-Message-ID-Hash
                         # headers which wer in the message contributing to the
                         # original_size, but weren't in the original text.
                         # Don't forget the space, delimeter, and newline!
                         len(self.text) + 50 + 52)

    def test_inject_text_with_keywords(self):
        # Keyword arguments are copied into the metadata.
        inject_text(self.mlist, self.text, foo='yes', bar='no')
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msgdata['foo'], 'yes')
        self.assertEqual(items[0].msgdata['bar'], 'no')

    def test_inject_message_id_hash(self):
        # When the injected message has a Message-ID header, the injected
        # message will also get a Message-ID-Hash header.
        inject_text(self.mlist, self.text)
        items = get_queue_messages('in', expected_count=1)
        self.assertEqual(items[0].msg['message-id-hash'],
                         'GUXXQKNCHBFQAHGBFMGCME6HKZCUUH3K')

    def test_inject_message_id_hash_without_message_id(self):
        # When the injected message does not have a Message-ID header, a
        # Message-ID header will be added, and the injected message will also
        # get a Message-ID-Hash header.
        filtered = self._remove_line('message-id')
        self.assertNotIn('Message-ID', filtered)
        self.assertNotIn('Message-ID-Hash', filtered)
        inject_text(self.mlist, filtered)
        items = get_queue_messages('in', expected_count=1)
        self.assertIn('message-id', items[0].msg)
        self.assertIn('message-id-hash', items[0].msg)

    def test_inject_text_non_ascii_string(self):
        # Test basic inject_text() call with a non-ascii string.
        inject_text(self.mlist, self.text2)
        items = get_queue_messages('in', expected_count=1)
        self.assertIsInstance(items[0].msg, Message)
        self.assertEqual(items[0].msg['message-id-hash'],
                         'GUXXQKNCHBFQAHGBFMGCME6HKZCUUH3K')
        # Remove the Message-ID-Hash header which isn't in the original text.
        del items[0].msg['message-id-hash']
        del items[0].msg['x-message-id-hash']
        self.assertMultiLineEqual(items[0].msg.as_bytes().decode('utf-8'),
                                  self.text2)

    def test_inject_text_non_ascii_bytes(self):
        # Test basic inject_text() call with non-ascii bytes.
        inject_text(self.mlist, self.text2.encode('utf-8'))
        items = get_queue_messages('in', expected_count=1)
        self.assertIsInstance(items[0].msg, Message)
        self.assertEqual(items[0].msg['message-id-hash'],
                         'GUXXQKNCHBFQAHGBFMGCME6HKZCUUH3K')
        # Remove the Message-ID-Hash header which isn't in the original text.
        del items[0].msg['message-id-hash']
        del items[0].msg['x-message-id-hash']
        self.assertMultiLineEqual(items[0].msg.as_bytes().decode('utf-8'),
                                  self.text2)



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""REST web service API context."""

from public import public
from zope.interface import Attribute, Interface


@public
class IAPI(Interface):
    """The REST web service context."""

    version = Attribute(
        """The REST API version as a string.""")

    version_info = Attribute(
        """The REST API version as a tuple of integers.""")

    def path_to(resource):
        """Return the full REST URL to the given resource.

        :param resource: Resource path string without the leading scheme,
            host, port, or API version information.
        :type resource: str
        :return: Full URL path to the resource, with the scheme, host, port
            and API version prepended.
        :rtype: str
        """

    def from_uuid(uuid):
        """Return the string representation of a UUID.

        :param uuid: The UUID to convert.
        :type uuid: UUID
        :return: The string representation of the UUID, as appropriate for the
            API version.  In 3.0 this is the representation of an integer,
            while in 3.1 it is the hex representation.
        :rtype: str
        """

    def to_uuid(uuid):
        """Return the UUID from the string representation.

        :param uuid: A UUID, or the string representation of the UUID.
        :type uuid: UUID or str
        :return: The UUID, converted if needed as appropriate for the
            API version.  In 3.0, the string representation is
            interpreted as an integer, while in 3.1 it is the hex
            string.
        :rtype: UUID
        """



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Interface for list storage, deleting, and finding."""

from mailman.interfaces.errors import MailmanError
from public import public
from zope.interface import Attribute, Interface


@public
class ListAlreadyExistsError(MailmanError):
    """Attempted to create a mailing list that already exists.

    Mailing list objects must be uniquely named by their fully qualified list
    name.
    """


@public
class NoSuchListError(MailmanError):
    """Attempt to access a mailing list that does not exist."""

    def __init__(self, fqdn_listname):
        self.fqdn_listname = fqdn_listname

    def __str__(self):
        return 'No such mailing list: {0.fqdn_listname}'.format(self)


@public
class ListCreatingEvent:
    """A mailing list is about to be created."""

    def __init__(self, fqdn_listname):
        self.fqdn_listname = fqdn_listname


@public
class ListCreatedEvent:
    """A mailing list was created."""

    def __init__(self, mlist):
        self.mailing_list = mlist


@public
class ListDeletingEvent:
    """A mailing list is about to be deleted."""

    def __init__(self, mailing_list):
        self.mailing_list = mailing_list


@public
class ListDeletedEvent:
    """A mailing list was deleted."""

    def __init__(self, fqdn_listname):
        self.fqdn_listname = fqdn_listname


@public
class IListManager(Interface):
    """The interface of the global list manager.

    The list manager manages `IMailingList` objects.  You can add and remove
    `IMailingList` objects from the list manager, and you can retrieve them
    from the manager via their fully qualified list name, e.g.:
    `mylist@example.com`.
    """

    def create(fqdn_listname):
        """Create a mailing list with the given name.

        :param fqdn_listname: The fully qualified name of the mailing list,
            e.g. `mylist@example.com`.
        :type fqdn_listname: Unicode
        :return: The newly created `IMailingList`.
        :raise `ListAlreadyExistsError` if the named list already exists.
        """

    def get(list_spec):
        """Return the mailing list with the given specification, if it exists.

        :param list_spec: Either the fully qualified name of the mailing list,
            or its List-ID.  If list_spec has an `@` in it, it's considered an
            FQDN listname, otherwise it's considered a List-ID.
        :type list_spec: str
        :return: the matching mailing list or None if no matching list is
            found.
        :rtype: IMailingList
        """

    def get_by_fqdn(fqdn_listname):
        """Return the mailing list with the given name, if it exists.

        :param fqdn_listname: The fully qualified name of the mailing list.
        :type fqdn_listname: str
        :return: the matching mailing list or None if no matching list is
            found.
        :rtype: IMailingList
        """

    def get_by_list_id(list_id):
        """Return the mailing list with the given list id, if it exists.

        :param list_id: The List-ID
        :type list_id: str
        :return: the matching mailing list or None if no matching list is
            found.
        :rtype: IMailingList
        """

    def delete(mlist):
        """Remove the mailing list from the database.

        :param mlist: The mailing list to delete.
        :type mlist: `IMailingList`
        """

    mailing_lists = Attribute(
        """An iterator over all the mailing list objects.

        The mailing lists are returned in order sorted by `list_id`.
        """)

    def __iter__():
        """An iterator over all the mailing lists.

        :return: iterator over `IMailingList`.
        """

    names = Attribute(
        """An iterator over the fully qualified list names of all mailing
        lists managed by this list manager.""")

    list_ids = Attribute(
        """An iterator over the list ids of all mailing lists managed by this
        list manager.""")

    name_components = Attribute(
        """An iterator over the 2-tuple of (list_name, mail_host) for all
        mailing lists managed by this list manager.""")

    def find(*, advertised=None, mail_host=None):
        """Search for mailing lists matching some criteria.

        The keyword arguments are mailing list properties that will be
        filtered upon.

        :return: The list of filtered mailing lists.
        :rtype: list of `IMailingList`
        """



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the Join and Leave commands."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.commands.eml_membership import Join, Leave
from mailman.email.message import Message
from mailman.interfaces.bans import IBanManager
from mailman.interfaces.mailinglist import SubscriptionPolicy
from mailman.interfaces.member import DeliveryMode
from mailman.interfaces.pending import IPendings
from mailman.interfaces.subscriptions import ISubscriptionManager
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.command import Results
from mailman.testing.helpers import set_preferred
from mailman.testing.layers import ConfigLayer
from zope.component import getUtility


class TestLeave(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._command = Leave()

    def test_confirm_leave_not_a_member(self):
        self._mlist.unsubscription_policy = SubscriptionPolicy.confirm
        # Try to unsubscribe someone who is not a member.  Anne is a real
        # user, with a validated address, but she is not a member of the
        # mailing list.
        anne = getUtility(IUserManager).create_user('anne@example.com')
        set_preferred(anne)
        # Initiate an unsubscription.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'leave: anne@example.com is not a member of ant@example.com')

    def test_leave_no_sender(self):
        # Initiate an unsubscription with no sender in the message.
        msg = Message()
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'leave: No valid email address found to unsubscribe')

    def test_leave_no_user(self):
        # Initiate an unsubscription for a non-user.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'No registered user for email address: anne@example.com')


class TestJoin(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('ant@example.com')
        self._command = Join()

    def test_join_successful(self):
        # Subscribe a member via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertIn('Confirmation email sent to anne@example.com',
                      str(results))

    def test_join_rfc2047_display(self):
        # Subscribe a member with RFC 2047 encoded display name via join.
        msg = Message()
        msg['From'] = '=?utf-8?q?Anne?= <anne@example.com>'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertIn('Confirmation email sent to Anne <anne@example.com>',
                      str(results))
        # Check the pending confirmation.
        pendings = list(getUtility(IPendings).find(self._mlist,
                                                   'subscription',
                                                   confirm=False))
        self.assertEqual(1, len(pendings))
        token = pendings[0][0]
        pended = getUtility(IPendings).confirm(token, expunge=False)
        self.assertEqual('Anne', pended['display_name'])
        self.assertEqual('anne@example.com', pended['email'])

    def test_join_rfc2047_display_with_comma(self):
        # Subscribe a member with RFC 2047 encoded display name containing a
        # comma and non-ascii via join.
        msg = Message()
        msg['From'] = '=?utf-8?q?J=C3=BCnk=2C_Anne?= <anne@example.com>'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertIn('Confirmation email sent to =?utf-8?b?SsO8bmssIEFubmU=?='
                      ' <anne@example.com>', str(results))
        # Check the pending confirmation.
        pendings = list(getUtility(IPendings).find(self._mlist,
                                                   'subscription',
                                                   confirm=False))
        self.assertEqual(1, len(pendings))
        token = pendings[0][0]
        pended = getUtility(IPendings).confirm(token, expunge=False)
        self.assertEqual('Jünk, Anne', pended['display_name'])
        self.assertEqual('anne@example.com', pended['email'])

    def test_join_digest(self):
        # Subscribe a member to digest via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, ('digest=mime',), results)
        self.assertIn('Confirmation email sent to anne@example.com',
                      str(results))

    def test_join_digest_works(self):
        # Subscribe a member to digest via join and verify it works.
        # Set things so the subscribe works now.
        self._mlist.subscription_policy = SubscriptionPolicy.open
        # Create a verified address for Anne.
        user = getUtility(IUserManager).make_user('anne@example.com', 'Anne')
        address = user.addresses[0]
        address.verified_on = address.registered_on
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, ('digest=mime',), results)
        # Anne is a member.
        members = list(self._mlist.members.members)
        self.assertEqual(1, len(members))
        self.assertEqual('anne@example.com', members[0].address.email)
        self.assertEqual(DeliveryMode.mime_digests, members[0].delivery_mode)

    def test_join_other(self):
        # Subscribe a different address via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {},
                              ('address=bob@example.com',), results)
        self.assertIn('Confirmation email sent to bob@example.com',
                      str(results))

    def test_join_other_bogus(self):
        # Try to subscribe a bogus different address via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {},
                              ('address=bogus',), results)
        self.assertIn('Invalid email address: bogus', str(results))

    def test_join_bad_argument(self):
        # Try to subscribe a member with a bad argument via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, ('digest=bogus',), results)
        self.assertIn('bad argument: digest=bogus', str(results))

    def test_join_bad_argument_name(self):
        # Try to subscribe a member with a bad argument via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, ('reg=bogus',), results)
        self.assertIn('bad argument: reg=bogus', str(results))

    def test_join_bad_argument_no_equal(self):
        # Try to subscribe a member with a bad argument via join.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, ('digest',), results)
        self.assertIn('bad argument: digest', str(results))

    def test_join_already_a_member(self):
        # Try to subscribe someone who is already a member.  Anne is a real
        # user, with a validated address, but she is not a member of the
        # mailing list yet.
        anne = getUtility(IUserManager).create_user('anne@example.com')
        set_preferred(anne)
        # First subscribe anne.
        ISubscriptionManager(self._mlist).register(anne, pre_verified=True,
                                                   pre_confirmed=True,
                                                   pre_approved=True)
        # Then initiate a subscription.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'anne@example.com is already a MemberRole.member of '
            'mailing list ant@example.com')

    def test_join_banned(self):
        # Try to subscribe someone who is banned.  Anne is a real
        # user, with a validated address, but she is not a member of the
        # mailing list and is banned from joining.
        # Add anne to the ban list.
        IBanManager(self._mlist).ban('anne@example.com')
        # Then initiate a subscription.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'anne@example.com is not allowed to subscribe to ant@example.com')

    def test_join_pending(self):
        self._mlist.subscription_policy = SubscriptionPolicy.confirm
        # Try to subscribe someone who already has a subscription pending.
        # Anne is a real user, with a validated address, who already has a
        # pending subscription for this mailing list.
        anne = getUtility(IUserManager).create_user('anne@example.com')
        set_preferred(anne)
        # Initiate a subscription.
        ISubscriptionManager(self._mlist).register(anne)
        # And try to subscribe.
        msg = Message()
        msg['From'] = 'anne@example.com'
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'anne@example.com has a pending subscription for ant@example.com')

    def test_join_posting_address(self):
        # Try to subscribe the list posting address.
        msg = Message()
        msg['From'] = self._mlist.posting_address
        results = Results()
        self._command.process(self._mlist, msg, {}, (), results)
        self.assertEqual(
            str(results).splitlines()[-1],
            'List posting address not allowed')



# Copyright (C) 2016-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the `max_size` rule."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.rules import max_size
from mailman.testing.helpers import specialized_message_from_string as mfs
from mailman.testing.layers import ConfigLayer


class TestMaximumSize(unittest.TestCase):
    """Test the max_size rule."""

    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')

    def test_max_size_returns_reason(self):
        # Ensure max_size rule returns a reason.
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A Subject
Message-ID: <ant>

A message body.
""")
        rule = max_size.MaximumSize()
        self._mlist.max_message_size = 1
        # Fake the size.
        msg.original_size = 2048
        msgdata = {}
        result = rule.check(self._mlist, msg, msgdata)
        self.assertTrue(result)
        self.assertEqual(msgdata['moderation_reasons'],
                         [('The message is larger than the {} KB maximum size',
                          1)])


class TestiFilteredMaximumSize(unittest.TestCase):
    """Test the max_size rule with filtered content."""

    layer = ConfigLayer

    def setUp(self):
        config.push('filter', """
        [mailman]
        check_max_size_on_filtered_message: yes
        """)
        self._mlist = create_list('test@example.com')
        self._mlist.max_message_size = 1
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: A Subject
Message-ID: <ant>
MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="AAAA"

--AAAA
Content-Type: text/plain

Small text plain part

--AAAA
Content-Type: application/octet-stream

This is a big part > 1K bytes
{}

--AAAA--
""".format(('x'*72 + '\n')*15))

    def test_no_filtering(self):
        # Test that unfiltered message is too big.
        self._mlist.filter_content = False
        rule = max_size.MaximumSize()
        msgdata = {}
        result = rule.check(self._mlist, self._msg, msgdata)
        self.assertTrue(result)

    def test_with_filtering(self):
        # Test that removing the big part passes the check.
        self._mlist.filter_content = True
        self._mlist.pass_types = ['multipart', 'text']
        rule = max_size.MaximumSize()
        msgdata = {}
        result = rule.check(self._mlist, self._msg, msgdata)
        self.assertFalse(result)

    def test_with_total_filtering(self):
        # Test that removing the whole message passes the check.
        self._mlist.filter_content = True
        self._mlist.pass_types = ['nothing']
        rule = max_size.MaximumSize()
        msgdata = {}
        result = rule.check(self._mlist, self._msg, msgdata)
        self.assertFalse(result)



# Copyright (C) 2007-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""The terminal 'accept' chain."""

import logging

from mailman.chains.base import TerminalChainBase
from mailman.config import config
from mailman.core.i18n import _
from mailman.interfaces.chain import AcceptEvent
from public import public
from zope.event import notify


log = logging.getLogger('mailman.vette')
SEMISPACE = '; '


@public
class AcceptChain(TerminalChainBase):
    """Accept the message for posting."""

    name = 'accept'
    description = _('Accept a message.')

    def _process(self, mlist, msg, msgdata):
        """See `TerminalChainBase`."""
        # Start by decorating the message with a header that contains a list
        # of all the rules that matched.  These metadata could be None or an
        # empty list.
        rule_hits = msgdata.get('rule_hits')
        if rule_hits:
            msg['X-Mailman-Rule-Hits'] = SEMISPACE.join(rule_hits)
        rule_misses = msgdata.get('rule_misses')
        if rule_misses:
            msg['X-Mailman-Rule-Misses'] = SEMISPACE.join(rule_misses)
        config.switchboards['pipeline'].enqueue(msg, msgdata)
        log.info('ACCEPT: %s', msg.get('message-id', 'n/a'))
        notify(AcceptEvent(mlist, msg, msgdata, self))



# Copyright (C) 2001-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Master subprocess watcher."""

import os
import sys
import click
import signal
import socket
import logging

from datetime import timedelta
from enum import Enum
from flufl.lock import Lock, NotLockedError, TimeOutError
from lazr.config import as_boolean
from mailman.config import config
from mailman.core.i18n import _
from mailman.core.initialize import initialize
from mailman.core.logging import reopen
from mailman.utilities.options import I18nCommand, validate_runner_spec
from mailman.version import MAILMAN_VERSION_FULL
from public import public


DOT = '.'
LOCK_LIFETIME = timedelta(days=1, hours=6)
SECONDS_IN_A_DAY = 86400
SUBPROC_START_WAIT = timedelta(seconds=20)

# Environment variables to forward into subprocesses.
PRESERVE_ENVS = (
    'COVERAGE_PROCESS_START',
    'LANG',
    'LANGUAGE',
    'LC_ADDRESS',
    'LC_ALL',
    'LC_COLLATE',
    'LC_CTYPE',
    'LC_IDENTIFICATION',
    'LC_MEASUREMENT',
    'LC_MESSAGES',
    'LC_MONETARY',
    'LC_NAME',
    'LC_NUMERIC',
    'LC_PAPER',
    'LC_TELEPHONE',
    'LC_TIME',
    'LOCALE_ARCHIVE',
    'MAILMAN_EXTRA_TESTING_CFG',
    'PYTHONPATH',
    'PYTHONHOME',
    )


@public
class WatcherState(Enum):
    """Enum for the state of the master process watcher."""
    # No lock has been acquired by any process.
    none = 0
    # Another master watcher is running.
    conflict = 1
    # No conflicting process exists.
    stale_lock = 2
    # Hostname from lock file doesn't match.
    host_mismatch = 3


@public
def master_state(lock_file=None):
    """Get the state of the master watcher.

    :param lock_file: Path to the lock file, otherwise `config.LOCK_FILE`.
    :type lock_file: str
    :return: 2-tuple of the WatcherState describing the state of the lock
        file, and the lock object.
    """
    if lock_file is None:
        lock_file = config.LOCK_FILE
    # We'll never acquire the lock, so the lifetime doesn't matter.
    lock = Lock(lock_file)
    try:
        hostname, pid, tempfile = lock.details
    except NotLockedError:
        return WatcherState.none, lock
    if hostname != socket.getfqdn():
        return WatcherState.host_mismatch, lock
    # Find out if the process exists by calling kill with a signal 0.
    try:
        os.kill(pid, 0)
        return WatcherState.conflict, lock
    except ProcessLookupError:
        # No matching process id.
        return WatcherState.stale_lock, lock


def acquire_lock_1(force, lock_file=None):
    """Try to acquire the master lock.

    :param force: Flag that controls whether to force acquisition of the lock.
    :type force: bool
    :param lock_file: Path to the lock file, otherwise `config.LOCK_FILE`.
    :type lock_file: str
    :return: The master lock.
    :raises: `TimeOutError` if the lock could not be acquired.
    """
    if lock_file is None:
        lock_file = config.LOCK_FILE
    lock = Lock(lock_file, LOCK_LIFETIME)
    try:
        lock.lock(timedelta(seconds=0.1))
        return lock
    except TimeOutError:
        if not force:
            raise
        # Force removal of lock first.
        hostname, pid, tempfile = lock.details
        os.unlink(lock_file)
        # Also remove any stale claim files.
        dname = os.path.dirname(lock_file)
        for fname in os.listdir(dname):
            fpath = os.path.join(dname, fname)
            if fpath.startswith(lock_file):
                os.unlink(fpath)
        return acquire_lock_1(force=False)


def acquire_lock(force):
    """Acquire the master lock.

    :param force: Flag that controls whether to force acquisition of the lock.
    :type force: bool
    :return: The master runner lock or None if the lock couldn't be acquired.
        In that case, an error messages is also printed to standard error.
    """
    try:
        lock = acquire_lock_1(force)
        return lock
    except TimeOutError:
        status, lock = master_state()
        if status is WatcherState.conflict:
            # Hostname matches and process exists.
            message = _("""\
The master lock could not be acquired because it appears as though another
master is already running.""")
        elif status is WatcherState.stale_lock:
            # Hostname matches but the process does not exist.
            program = sys.argv[0]                   # noqa: F841
            message = _("""\
The master lock could not be acquired.  It appears as though there is a stale
master lock.  Try re-running ${program} with the --force flag.""")
        elif status is WatcherState.host_mismatch:
            # Hostname doesn't even match.
            hostname, pid, tempfile = lock.details
            message = _("""\
The master lock could not be acquired, because it appears as if some process
on some other host may have acquired it.  We can't test for stale locks across
host boundaries, so you'll have to clean this up manually.

Lock file: ${config.LOCK_FILE}
Lock host: ${hostname}

Exiting.""")
        else:
            assert status is WatcherState.none, (
                'Invalid enum value: {}'.format(status))
            hostname, pid, tempfile = lock.details
            message = _("""\
For unknown reasons, the master lock could not be acquired.

Lock file: ${config.LOCK_FILE}
Lock host: ${hostname}

Exiting.""")
        print(message, file=sys.stderr)
        sys.exit(1)


class PIDWatcher:
    """A class which safely manages child process ids."""

    def __init__(self):
        self._pids = {}

    def __contains__(self, pid):
        return pid in self._pids.keys()

    def __iter__(self):
        # Safely iterate over all the keys in the dictionary.  Because
        # asynchronous signals are involved, the dictionary's size could
        # change during iteration.  Iterate over a copy of the keys to avoid
        # that.
        for pid in list(self._pids):
            yield pid

    def add(self, pid, info):
        """Add process information.

        :param pid: The process id.  The watcher must not already be tracking
            this process id.
        :type pid: int
        :param info: The process information.
        :type info: 4-tuple consisting of
            (runner-name, slice-number, slice-count, restart-count)
        """
        old_info = self._pids.get(pid)
        assert old_info is None, (
            'Duplicate process id {0} with existing info: {1}'.format(
                pid, old_info))
        self._pids[pid] = info

    def pop(self, pid):
        """Remove and return existing process information.

        :param pid: The process id.  The watcher must already be tracking this
            process id.
        :type pid: int
        :return: The process information.
        :rtype: 4-tuple consisting of
            (runner-name, slice-number, slice-count, restart-count)
        :raise KeyError: if the process id is not being tracked.
        """
        return self._pids.pop(pid)

    def drop(self, pid):
        """Remove and return existing process information.

        This is like `pop()` except that no `KeyError` is raised if the
        process id is not being tracked.

        :param pid: The process id.
        :type pid: int
        :return: The process information, or None if the process id is not
            being tracked.
        :rtype: 4-tuple consisting of
            (runner-name, slice-number, slice-count, restart-count)
        """
        return self._pids.pop(pid, None)


@public
class Loop:
    """Main control loop class."""

    def __init__(self, lock=None, restartable=None, config_file=None):
        self._lock = lock
        self._restartable = restartable
        self._config_file = config_file
        self._kids = PIDWatcher()

    def install_signal_handlers(self):
        """Install various signals handlers for control from the master."""
        log = logging.getLogger('mailman.runner')
        # Set up our signal handlers.  Also set up a SIGALRM handler to
        # refresh the lock once per day.  The lock lifetime is 1 day + 6 hours
        # so this should be plenty.
        def sigalrm_handler(signum, frame):                      # noqa: E306
            self._lock.refresh()
            signal.alarm(SECONDS_IN_A_DAY)
        signal.signal(signal.SIGALRM, sigalrm_handler)
        signal.alarm(SECONDS_IN_A_DAY)
        # SIGHUP tells the runners to close and reopen their log files.
        def sighup_handler(signum, frame):                        # noqa: E306
            reopen()
            for pid in self._kids:
                os.kill(pid, signal.SIGHUP)
            log.info('Master watcher caught SIGHUP.  Re-opening log files.')
        signal.signal(signal.SIGHUP, sighup_handler)
        # SIGUSR1 is used by 'mailman restart'.
        def sigusr1_handler(signum, frame):                       # noqa: E306
            for pid in self._kids:
                os.kill(pid, signal.SIGUSR1)
            log.info('Master watcher caught SIGUSR1.  Exiting.')
        signal.signal(signal.SIGUSR1, sigusr1_handler)
        # SIGTERM is what init will kill this process with when changing run
        # levels.  It's also the signal 'mailman stop' uses.
        def sigterm_handler(signum, frame):                       # noqa: E306
            for pid in self._kids:
                os.kill(pid, signal.SIGTERM)
            log.info('Master watcher caught SIGTERM.  Exiting.')
        signal.signal(signal.SIGTERM, sigterm_handler)
        # SIGINT is what control-C gives.
        def sigint_handler(signum, frame):                        # noqa: E306
            for pid in self._kids:
                os.kill(pid, signal.SIGINT)
            log.info('Master watcher caught SIGINT.  Restarting.')
        signal.signal(signal.SIGINT, sigint_handler)

    def _start_runner(self, spec):
        """Start a runner.

        All arguments are passed to the process.

        :param spec: A runner spec, in a format acceptable to
            bin/runner's --runner argument, e.g. name:slice:count
        :type spec: string
        :return: The process id of the child runner.
        :rtype: int
        """
        pid = os.fork()
        if pid:
            # Parent.
            return pid
        # Child.
        #
        # Set the environment variable which tells the runner that it's
        # running under bin/master control.  This subtly changes the error
        # behavior of bin/runner.
        env = {'MAILMAN_UNDER_MASTER_CONTROL': '1'}
        # Craft the command line arguments for the exec() call.
        rswitch = '--runner=' + spec
        # Always pass the explicit path to the configuration file to the
        # sub-runners.  This avoids any debate about which cfg file is used.
        config_file = (config.filename if self._config_file is None
                       else self._config_file)
        # Wherever master lives, so too must live the runner script.
        exe = os.path.join(config.BIN_DIR, 'runner')   # pragma: nocover
        # config.PYTHON, which is the absolute path to the Python interpreter,
        # must be given as argv[0] due to Python's library search algorithm.
        args = [sys.executable, sys.executable, exe,   # pragma: nocover
                '-C', config_file, rswitch]
        log = logging.getLogger('mailman.runner')
        log.debug('starting: %s', args)
        # We must pass this environment variable through if it's set,
        # otherwise runner processes will not have the correct VAR_DIR.
        var_dir = os.environ.get('MAILMAN_VAR_DIR')
        if var_dir is not None:
            env['MAILMAN_VAR_DIR'] = var_dir
        # For the testing framework, if these environment variables are set,
        # pass them on to the subprocess.
        for envvar in PRESERVE_ENVS:
            if envvar in os.environ:
                env[envvar] = os.environ[envvar]
        args.append(env)
        os.execle(*args)
        # We should never get here.
        raise RuntimeError('os.execle() failed')

    def start_runners(self, runner_names=None):
        """Start all the configured runners.

        :param runners: If given, a sequence of runner names to start.  If not
            given, this sequence is taken from the configuration file.
        :type runners: a sequence of strings
        """
        if not runner_names:
            runner_names = []
            for runner_config in config.runner_configs:
                # Strip off the 'runner.' prefix.
                assert runner_config.name.startswith('runner.'), (
                    'Unexpected runner configuration section name: {}'.format(
                        runner_config.name))
                runner_names.append(runner_config.name[7:])
        # For each runner we want to start, find their config section, which
        # will tell us the name of the class to instantiate, along with the
        # number of hash space slices to manage.
        for name in runner_names:
            section_name = 'runner.' + name
            # Let AttributeError propagate.
            runner_config = getattr(config, section_name)
            if not as_boolean(runner_config.start):
                continue
            # Find out how many runners to instantiate.  This must be a power
            # of 2.
            count = int(runner_config.instances)
            assert (count & (count - 1)) == 0, (
                'Runner "{0}", not a power of 2: {1}'.format(name, count))
            for slice_number in range(count):
                # runner name, slice #, # of slices, restart count
                info = (name, slice_number, count, 0)
                spec = '{0}:{1:d}:{2:d}'.format(name, slice_number, count)
                pid = self._start_runner(spec)
                log = logging.getLogger('mailman.runner')
                log.debug('[{0:d}] {1}'.format(pid, spec))
                self._kids.add(pid, info)

    def _pause(self):
        """Sleep until a signal is received."""
        # Sleep until a signal is received.  This prevents the master from
        # exiting immediately even if there are no runners (as happens in the
        # test suite).
        signal.pause()

    def loop(self):
        """Main loop.

        Wait until all the runner subprocesses have exited, restarting them if
        necessary and configured to do so.
        """
        log = logging.getLogger('mailman.runner')
        log.info('Master started')
        self._pause()
        while True:
            try:
                pid, status = os.wait()
            except ChildProcessError:
                # No children?  We're done.
                break
            except InterruptedError:                         # pragma: nocover
                # If the system call got interrupted, just restart it.
                continue
            if pid not in self._kids:                        # pragma: nocover
                # This is not a runner subprocess that we own.  E.g. maybe a
                # plugin started it.
                continue
            # Find out why the subprocess exited by getting the signal
            # received or exit status.
            if os.WIFSIGNALED(status):
                why = os.WTERMSIG(status)                    # pragma: nocover
                sig_or_exit = 'SIGNAL '                      # pragma: nocover
            elif os.WIFEXITED(status):
                why = os.WEXITSTATUS(status)
                sig_or_exit = 'EXIT '
            else:                                            # pragma: nocover
                why = None
                sig_or_exit = 'UNKNOWN'
            # We'll restart the subprocess if it exited with a SIGUSR1 or
            # because of a failure (i.e. no exit signal), and the no-restart
            # command line switch was not given.  This lets us better handle
            # runaway restarts (e.g.  if the subprocess had a syntax error!)
            rname, slice_number, count, restarts = self._kids.pop(pid)
            config_name = 'runner.' + rname
            restart = False
            if why == signal.SIGUSR1 and self._restartable:  # pragma: nocover
                restart = True
            # Have we hit the maximum number of restarts?
            restarts += 1
            max_restarts = int(getattr(config, config_name).max_restarts)
            if restarts > max_restarts:
                restart = False
            # Are we permanently non-restartable?
            log.debug("""\
Master detected subprocess exit
(pid: {0:d}, why: {1}{2}, class: {3}, slice: {4:d}/{5:d}) {6}""".format(
                     pid, sig_or_exit, why, rname, slice_number + 1, count,
                     ('[restarting]' if restart else '')))
            # See if we've reached the maximum number of allowable restarts.
            if restarts > max_restarts:
                log.info("""\
Runner {0} reached maximum restart limit of {1:d}, not restarting.""".format(
                         rname, max_restarts))
            # Now perhaps restart the process unless it exited with a
            # SIGTERM or we aren't restarting.
            if restart:
                spec = '{0}:{1:d}:{2:d}'.format(rname, slice_number, count)
                new_pid = self._start_runner(spec)
                new_info = (rname, slice_number, count, restarts)
                self._kids.add(new_pid, new_info)
        log.info('Master stopped')

    def cleanup(self):
        """Ensure that all children have exited."""
        log = logging.getLogger('mailman.runner')
        # Send SIGTERMs to all the child processes and wait for them all to
        # exit.
        for pid in self._kids:
            try:
                os.kill(pid, signal.SIGTERM)
            except ProcessLookupError:              # pragma: nocover
                # The child has already exited.
                log.info('ESRCH on pid: %d', pid)
            except OSError:                         # pragma: nocover
                # XXX I'm not so sure about this.  It preserves the semantics
                # before conversion to PEP 3151 exceptions.  But is it right?
                pass
        # Wait for all the children to go away.
        while self._kids:
            try:
                pid, status = os.wait()
                self._kids.drop(pid)
            except ChildProcessError:
                break
            except InterruptedError:                # pragma: nocover
                continue


@click.command(
    cls=I18nCommand,
    context_settings=dict(help_option_names=['-h', '--help']),
    help=_("""\
    Master subprocess watcher.

    Start and watch the configured runners, ensuring that they stay alive and
    kicking.  Each runner is forked and exec'd in turn, with the master waiting
    on their process ids.  When it detects a child runner has exited, it may
    restart it.

    The runners respond to SIGINT, SIGTERM, SIGUSR1 and SIGHUP.  SIGINT,
    SIGTERM and SIGUSR1 all cause a runner to exit cleanly.  The master will
    restart runners that have exited due to a SIGUSR1 or some kind of other
    exit condition (say because of an uncaught exception).  SIGHUP causes the
    master and the runners to close their log files, and reopen then upon the
    next printed message.

    The master also responds to SIGINT, SIGTERM, SIGUSR1 and SIGHUP, which it
    simply passes on to the runners.  Note that the master will close and
    reopen its own log files on receipt of a SIGHUP.  The master also leaves
    its own process id in the file specified in the configuration file but you
    normally don't need to use this PID directly."""))
@click.option(
    '-C', '--config', 'config_file',
    envvar='MAILMAN_CONFIG_FILE',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help=_("""\
    Configuration file to use.  If not given, the environment variable
    MAILMAN_CONFIG_FILE is consulted and used if set.  If neither are given, a
    default configuration file is loaded."""))
@click.option(
    '--no-restart', '-n', 'restartable',
    is_flag=True, default=True,
    help=_("""\
    Don't restart the runners when they exit because of an error or a SIGUSR1.
    Use this only for debugging."""))
@click.option(
    '--force', '-f',
    is_flag=True, default=False,
    help=_("""\
    If the master watcher finds an existing master lock, it will normally exit
    with an error message.  With this option,the master will perform an extra
    level of checking.  If a process matching the host/pid described in the
    lock file is running, the master will still exit, requiring you to manually
    clean up the lock.  But if no matching process is found, the master will
    remove the apparently stale lock and make another attempt to claim the
    master lock."""))
@click.option(
    '--runners', '-r',
    metavar='runner[:slice:range]',
    callback=validate_runner_spec, default=None,
    multiple=True,
    help=_("""\
    Override the default set of runners that the master will invoke, which is
    typically defined in the configuration file.  Multiple -r options may be
    given.  The values for -r are passed straight through to bin/runner."""))
@click.option(
    '-v', '--verbose',
    is_flag=True, default=False,
    help=_('Display more debugging information to the log file.'))
@click.version_option(MAILMAN_VERSION_FULL)
@public
def main(config_file, restartable, force, runners, verbose):
    # XXX https://github.com/pallets/click/issues/303
    """Master subprocess watcher.

    Start and watch the configured runners and ensure that they stay
    alive and kicking.  Each runner is forked and exec'd in turn, with
    the master waiting on their process ids.  When it detects a child
    runner has exited, it may restart it.

    The runners respond to SIGINT, SIGTERM, SIGUSR1 and SIGHUP.  SIGINT,
    SIGTERM and SIGUSR1 all cause a runner to exit cleanly.  The master
    will restart runners that have exited due to a SIGUSR1 or some kind
    of other exit condition (say because of an uncaught exception).
    SIGHUP causes the master and the runners to close their log files,
    and reopen then upon the next printed message.

    The master also responds to SIGINT, SIGTERM, SIGUSR1 and SIGHUP,
    which it simply passes on to the runners.  Note that the master will
    close and reopen its own log files on receipt of a SIGHUP.  The
    master also leaves its own process id in the file `data/master.pid`
    but you normally don't need to use this pid directly.
    """
    initialize(config_file, verbose)
    # Acquire the master lock, exiting if we can't.  We'll let the caller
    # handle any clean up or lock breaking.  No `with` statement here because
    # Lock's constructor doesn't support a timeout.
    lock = acquire_lock(force)
    try:
        with open(config.PID_FILE, 'w') as fp:
            print(os.getpid(), file=fp)
        loop = Loop(lock, restartable, config.filename)
        loop.install_signal_handlers()
        try:
            loop.start_runners(runners)
            loop.loop()
        finally:
            loop.cleanup()
            os.remove(config.PID_FILE)
    finally:
        lock.unlock()



# Copyright (C) 2014-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Test the filter handler."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.config import config
from mailman.interfaces.mime import FilterAction
from mailman.interfaces.pipeline import DiscardMessage
from mailman.testing.helpers import specialized_message_from_string as mfs
from mailman.testing.layers import ConfigLayer


class TestFilters(unittest.TestCase):
    layer = ConfigLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')

    def test_discard_when_outer_type_matches(self):
        # When the outer MIME type of the message matches a filter type, the
        # entire message is discarded.
        self._mlist.filter_content = True
        self._mlist.filter_types = ['image/jpeg']
        self._mlist.filter_action = FilterAction.discard
        msg = mfs("""\
From: aperson@example.com
Content-Type: image/jpeg
MIME-Version: 1.0

xxxxx
""")
        self.assertRaises(DiscardMessage,
                          config.handlers['mime-delete'].process,
                          self._mlist, msg, {})



# Copyright (C) 2011-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""Moderation tests."""

import unittest

from mailman.app.lifecycle import create_list
from mailman.app.moderator import (
    handle_message,
    handle_unsubscription,
    hold_message,
    hold_unsubscription,
)
from mailman.chains.hold import HeldMessagePendable
from mailman.interfaces.action import Action
from mailman.interfaces.member import MemberRole
from mailman.interfaces.messages import IMessageStore
from mailman.interfaces.pending import IPendings
from mailman.interfaces.requests import IListRequests
from mailman.interfaces.subscriptions import ISubscriptionManager
from mailman.interfaces.usermanager import IUserManager
from mailman.runners.incoming import IncomingRunner
from mailman.runners.outgoing import OutgoingRunner
from mailman.runners.pipeline import PipelineRunner
from mailman.testing.helpers import (
    get_queue_messages,
    make_testable_runner,
    set_preferred,
    specialized_message_from_string as mfs,
)
from mailman.testing.layers import SMTPLayer
from mailman.utilities.datetime import now
from zope.component import getUtility


class TestModeration(unittest.TestCase):
    """Test moderation functionality."""

    layer = SMTPLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._request_db = IListRequests(self._mlist)
        self._msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: hold me
Message-ID: <alpha>

""")
        self._in = make_testable_runner(IncomingRunner, 'in')
        self._pipeline = make_testable_runner(PipelineRunner, 'pipeline')
        self._out = make_testable_runner(OutgoingRunner, 'out')

    def test_accepted_message_gets_posted(self):
        # A message that is accepted by the moderator should get posted to the
        # mailing list.  LP: #827697
        msgdata = dict(listname='test@example.com',
                       recipients=['bart@example.com'])
        request_id = hold_message(self._mlist, self._msg, msgdata)
        handle_message(self._mlist, request_id, Action.accept)
        self._in.run()
        self._pipeline.run()
        self._out.run()
        messages = list(SMTPLayer.smtpd.messages)
        self.assertEqual(len(messages), 1)
        message = messages[0]
        # We don't need to test the entire posted message, just the bits that
        # prove it got sent out.
        self.assertIn('x-mailman-version', message)
        self.assertIn('x-peer', message)
        # The X-Mailman-Approved-At header has local timezone information in
        # it, so test that separately.
        self.assertEqual(message['x-mailman-approved-at'][:-5],
                         'Mon, 01 Aug 2005 07:49:23 ')
        del message['x-mailman-approved-at']
        # The Message-ID matches the original.
        self.assertEqual(message['message-id'], '<alpha>')
        # Anne sent the message and the mailing list received it.
        self.assertEqual(message['from'], 'anne@example.com')
        self.assertEqual(message['to'], 'test@example.com')
        # The Subject header has the list's prefix.
        self.assertEqual(message['subject'], '[Test] hold me')
        # The list's -bounce address is the actual sender, and Bart is the
        # only actual recipient.  These headers are added by the testing
        # framework and don't show up in production.  They match the RFC 5321
        # envelope.
        self.assertEqual(message['x-mailfrom'], 'test-bounces@example.com')
        self.assertEqual(message['x-rcptto'], 'bart@example.com')

    def test_missing_accepted_message_gets_posted(self):
        # A message that is accepted by the moderator should get posted to the
        # mailing list.  LP: #827697
        msgdata = dict(listname='test@example.com',
                       recipients=['bart@example.com'])
        request_id = hold_message(self._mlist, self._msg, msgdata)
        message_store = getUtility(IMessageStore)
        message_store.delete_message('<alpha>')
        handle_message(self._mlist, request_id, Action.accept)
        self._in.run()
        self._pipeline.run()
        self._out.run()
        messages = list(SMTPLayer.smtpd.messages)
        self.assertEqual(len(messages), 1)
        message = messages[0]
        # We don't need to test the entire posted message, just the bits that
        # prove it got sent out.
        self.assertIn('x-mailman-version', message)
        self.assertIn('x-peer', message)
        # The X-Mailman-Approved-At header has local timezone information in
        # it, so test that separately.
        self.assertEqual(message['x-mailman-approved-at'][:-5],
                         'Mon, 01 Aug 2005 07:49:23 ')
        del message['x-mailman-approved-at']
        # The Message-ID matches the original.
        self.assertEqual(message['message-id'], '<alpha>')
        # Anne sent the message and the mailing list received it.
        self.assertEqual(message['from'], 'anne@example.com')
        self.assertEqual(message['to'], 'test@example.com')
        # The Subject header has the list's prefix.
        self.assertEqual(message['subject'], '[Test] hold me')
        # The list's -bounce address is the actual sender, and Bart is the
        # only actual recipient.  These headers are added by the testing
        # framework and don't show up in production.  They match the RFC 5321
        # envelope.
        self.assertEqual(message['x-mailfrom'], 'test-bounces@example.com')
        self.assertEqual(message['x-rcptto'], 'bart@example.com')

    def test_hold_action_alias_for_defer(self):
        # In handle_message(), the 'hold' action is the same as 'defer' for
        # purposes of this API.
        request_id = hold_message(self._mlist, self._msg)
        handle_message(self._mlist, request_id, Action.defer)
        # The message is still in the pending requests.
        key, data = self._request_db.get_request(request_id)
        self.assertEqual(key, '<alpha>')
        handle_message(self._mlist, request_id, Action.hold)
        key, data = self._request_db.get_request(request_id)
        self.assertEqual(key, '<alpha>')

    def test_lp_1031391(self):
        # LP: #1031391 msgdata['received_time'] gets added by the LMTP server.
        # The value is a datetime.  If this message gets held, it will break
        # pending requests since they require string keys and values.
        received_time = now()
        msgdata = dict(received_time=received_time)
        request_id = hold_message(self._mlist, self._msg, msgdata)
        key, data = self._request_db.get_request(request_id)
        self.assertEqual(data['received_time'], received_time)

    def test_forward(self):
        # We can forward the message to an email address.
        request_id = hold_message(self._mlist, self._msg)
        handle_message(self._mlist, request_id, Action.discard,
                       forward=['zack@example.com'])
        # The forwarded message lives in the virgin queue.
        items = get_queue_messages('virgin', expected_count=1)
        self.assertEqual(str(items[0].msg['subject']),
                         'Forward of moderated message')
        self.assertEqual(list(items[0].msgdata['recipients']),
                         ['zack@example.com'])

    def test_missing_forward(self):
        # We can forward the message to an email address.
        request_id = hold_message(self._mlist, self._msg)
        message_store = getUtility(IMessageStore)
        message_store.delete_message('<alpha>')
        handle_message(self._mlist, request_id, Action.discard,
                       forward=['zack@example.com'])
        # The forwarded message lives in the virgin queue.
        items = get_queue_messages('virgin', expected_count=1)
        self.assertEqual(str(items[0].msg['subject']),
                         'Forward of moderated message')
        self.assertEqual(list(items[0].msgdata['recipients']),
                         ['zack@example.com'])
        payload = items[0].msg.get_payload()[0]
        self.assertIsNotNone(payload)
        self.assertEqual('<alpha>', payload['message-id'])

    def test_survive_a_deleted_message(self):
        # When the message that should be deleted is not found in the store,
        # no error is raised.
        request_id = hold_message(self._mlist, self._msg)
        message_store = getUtility(IMessageStore)
        message_store.delete_message('<alpha>')
        handle_message(self._mlist, request_id, Action.discard)
        self.assertEqual(self._request_db.count, 0)

    def test_handled_message_removed_from_store(self):
        # The message is removed from the store and the pendings db when it's
        # been disposed of.
        request_id = hold_message(self._mlist, self._msg)
        # Get the hash for this pending request.
        hash = list(self._request_db.held_requests)[0].data_hash
        handle_message(self._mlist, request_id, Action.discard)
        self.assertEqual(self._request_db.count, 0)
        message = getUtility(IMessageStore).get_message_by_id('<alpha>')
        self.assertIsNone(message)
        self.assertIsNone(getUtility(IPendings).confirm(hash))

    def test_handled_cross_posted_message_not_removed(self):
        # A cross posted message is not removed when handled on the first list.
        mlist2 = create_list('test2@example.com')
        request_db2 = IListRequests(mlist2)
        request_id = hold_message(self._mlist, self._msg)
        request_id2 = hold_message(mlist2, self._msg)
        # Get the hashes for these pending requests.
        hash0 = list(self._request_db.held_requests)[0].data_hash
        hash1 = list(request_db2.held_requests)[0].data_hash
        # Handle the first list's message.
        handle_message(self._mlist, request_id, Action.discard)
        # There's now only the request for list2.
        self.assertEqual(self._request_db.count, 0)
        self.assertEqual(request_db2.count, 1)
        message = getUtility(IMessageStore).get_message_by_id('<alpha>')
        self.assertIsNotNone(message)
        self.assertIsNone(getUtility(IPendings).confirm(hash0))
        self.assertIsNotNone(getUtility(IPendings).confirm(hash1,
                                                           expunge=False))
        # Handle the second list's message.
        handle_message(mlist2, request_id2, Action.discard)
        # Now the request and message are gone.
        self.assertEqual(request_db2.count, 0)
        message = getUtility(IMessageStore).get_message_by_id('<alpha>')
        self.assertIsNone(message)
        self.assertIsNone(getUtility(IPendings).confirm(hash1))

    def test_all_pendings_removed(self):
        # A held message pends two tokens, One for the moderator and one for
        # the user.  Ensure both are removed when message is handled.
        request_id = hold_message(self._mlist, self._msg)
        # The hold chain does more.
        pendings = getUtility(IPendings)
        user_hash = pendings.add(HeldMessagePendable(
            id=request_id, list_id=self._mlist.list_id))
        # Get the hash for this pending request.
        hash = list(self._request_db.held_requests)[0].data_hash
        handle_message(self._mlist, request_id, Action.discard)
        self.assertEqual(self._request_db.count, 0)
        self.assertIsNone(pendings.confirm(hash))
        self.assertIsNone(pendings.confirm(user_hash))

    def test_all_pendings_removed_old(self):
        # This is different from the testcase above in that it
        # handles the old style HeldMessagePendable where the
        # list_id is not added to pendedkeyvalue. It is only
        # meant so that new versions can handle old pendables.
        request_id = hold_message(self._mlist, self._msg)
        # The hold chain does more.
        pendings = getUtility(IPendings)
        user_hash = pendings.add(HeldMessagePendable(id=request_id))
        # Get the hash for this pending request.
        hash = list(self._request_db.held_requests)[0].data_hash
        handle_message(self._mlist, request_id, Action.discard)
        self.assertEqual(self._request_db.count, 0)
        self.assertIsNone(pendings.confirm(hash))
        self.assertIsNone(pendings.confirm(user_hash))

    def test_rejection_subject_decoded(self):
        # Test that a rejected message with an RFC 2047 encoded Subject: has
        # the Subject: decoded for the rejection
        del self._msg['subject']
        self._msg['Subject'] = '=?utf-8?q?hold_me?='
        request_id = hold_message(self._mlist, self._msg)
        handle_message(self._mlist, request_id, Action.reject)
        # The rejected message lives in the virgin queue.
        items = get_queue_messages('virgin', expected_count=1)
        rejection = items[0].msg
        self.assertEqual(str(rejection['subject']),
                         'Request to mailing list "Test" rejected')
        self.assertIn('Posting of your message titled "hold me"',
                      rejection.get_payload())


class TestUnsubscription(unittest.TestCase):
    """Test unsubscription requests."""

    layer = SMTPLayer

    def setUp(self):
        self._mlist = create_list('test@example.com')
        self._manager = ISubscriptionManager(self._mlist)

    def test_unsubscribe_defer(self):
        # When unsubscriptions must be approved by the moderator, but the
        # moderator defers this decision.
        user_manager = getUtility(IUserManager)
        anne = user_manager.create_address('anne@example.org', 'Anne Person')
        token, token_owner, member = self._manager.register(
            anne, pre_verified=True, pre_confirmed=True, pre_approved=True)
        self.assertIsNone(token)
        self.assertEqual(member.address.email, 'anne@example.org')
        bart = user_manager.create_user('bart@example.com', 'Bart User')
        address = set_preferred(bart)
        self._mlist.subscribe(address, MemberRole.moderator)
        # Now hold and handle an unsubscription request.
        token = hold_unsubscription(self._mlist, 'anne@example.org')
        handle_unsubscription(self._mlist, token, Action.defer)
        items = get_queue_messages('virgin', expected_count=2)
        # Find the moderator message.
        for item in items:
            if item.msg['to'] == 'test-owner@example.com':
                break
        else:
            raise AssertionError('No moderator email found')
        self.assertEqual(
            item.msgdata['recipients'], {'test-owner@example.com'})
        self.assertEqual(
            item.msg['subject'],
            'New unsubscription request from Test by anne@example.org')

    def test_bogus_token(self):
        # Try to handle an unsubscription with a bogus token.
        self.assertRaises(LookupError, self._manager.confirm, None)



#! /usr/bin/env python3
# Author: Abhilash Raj
# Date: Sept 26 2020
#
# The purpose of this script is to generate a multi-lingual en po file for
# Mailman Core. Multi-lingual PO files use both a .pot file and a reference po
# file. The reference in our case is en and is what is displayed in Weblate
# when translating.
#
# Since there are no utilities to create multi-lingual reference PO file, this
# script exists to solve that purpose. Here is how it works:
#
# It will read the default en po file and set all msgstr to be same as
# msgid. The only exception to this is email templates, where the msgid is the
# filename of the template and msgstr is the content of that file.

from pathlib import Path


try:
    from babel.messages.pofile import read_po, write_po
except ImportError:
    print('Please install `babel` to run this script.')
    exit(1)

PO_PATH = Path('src/mailman/messages/en/LC_MESSAGES/mailman.po')
TEMPLATE_BASE_PATH = Path('src/mailman/templates/en')


def get_po(path):
    "Read the po file path and return a Catalog object."
    with path.open() as fd:
        catalog = read_po(fd)
    return catalog

def put_po(path, catalog):
    "Write the catalog object to the po file path."
    with path.open('bw') as fd:
        write_po(fd, catalog, include_previous=True, width=85, )

def get_template(name):
    "Get the template text with the name if it exists."
    template_path = TEMPLATE_BASE_PATH / name
    if not template_path.exists():
        return ''
    return template_path.read_text().rstrip('\n')


def main():
    catalog = get_po(PO_PATH)
    for each in catalog:
        if each.id.endswith('.txt'):
            each.string = get_template(each.id)
        else:
            each.string = each.id
    put_po(PO_PATH, catalog)
    print(f'Updated {PO_PATH}')

main()



# Copyright (C) 2009-2023 by the Free Software Foundation, Inc.
#
# This file is part of GNU Mailman.
#
# GNU Mailman is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# GNU Mailman is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# GNU Mailman.  If not, see <https://www.gnu.org/licenses/>.

"""VERP delivery."""

import logging

from mailman.config import config
from mailman.mta.base import IndividualDelivery
from mailman.utilities.email import split_email
from mailman.utilities.string import expand
from public import public


DOT = '.'
log = logging.getLogger('mailman.smtp')


@public
class VERPMixin:
    """Mixin for VERP functionality.

    This works by overriding the base class's _get_sender() method to return
    the VERP'd envelope sender.  It expects the individual recipient's address
    to be squirreled away in the message metadata.
    """
    def _get_sender(self, mlist, msg, msgdata):
        """Return the recipient's address VERP encoded in the sender.

        :param mlist: The mailing list being delivered to.
        :type mlist: `IMailingList`
        :param msg: The original message being delivered.
        :type msg: `Message`
        :param msgdata: Additional message metadata for this delivery.
        :type msgdata: dictionary
        """
        sender = super()._get_sender(mlist, msg, msgdata)
        if msgdata.get('verp', False):
            log.debug('VERPing %s', msg.get('message-id'))
            recipient = msgdata['recipient']
            sender_mailbox, sender_domain = split_email(sender)
            # Encode the recipient's address for VERP.
            recipient_mailbox, recipient_domain = split_email(recipient)
            if recipient_domain is None:
                # The recipient address is not fully-qualified.  We can't
                # deliver it to this person, nor can we craft a valid verp
                # header.  I don't think there's much we can do except ignore
                # this recipient.
                log.info('Skipping VERP delivery to unqual recip: %s',
                         recipient)
                return sender
            return '{0}@{1}'.format(
                expand(config.mta.verp_format, mlist, dict(
                    bounces=sender_mailbox,
                    local=recipient_mailbox,
                    domain=DOT.join(recipient_domain))),
                DOT.join(sender_domain))
        else:
            return sender

    def avoid_duplicates(self, mlist, msg, msgdata):
        """Flag the message for duplicate avoidance.

        We can flag the mail as a duplicate for each member, if they've
        already received this message, as calculated by Message-ID.  See
        `AvoidDuplicates.py`_ for details.
        """
        recipient = msgdata['recipient']
        del msg['x-mailman-copy']
        if recipient in msgdata.get('add-dup-header', {}):
            msg['X-Mailman-Copy'] = 'yes'


@public
class VERPDelivery(VERPMixin, IndividualDelivery):
    """Deliver a unique message to the MSA for each recipient."""

    def __init__(self):
        """See `IndividualDelivery`."""
        super().__init__()
        self.callbacks.append(self.avoid_duplicates)



"""add_alias_domain

Revision ID: dfe82cf73702
Revises: fa0d96e28631
Create Date: 2016-10-07 16:50:53.368932

"""

import sqlalchemy as sa

from alembic import op
from mailman.database.helpers import exists_in_db
from mailman.database.types import SAUnicode


# revision identifiers, used by Alembic.
revision = 'dfe82cf73702'
down_revision = '3f31035ed0d7'


def upgrade():
    if not exists_in_db(op.get_bind(), 'domain', 'alias_domain'):
        op.add_column(
            'domain', sa.Column('alias_domain', SAUnicode, nullable=True)
            )


def downgrade():
    with op.batch_alter_table('domain') as batch_op:
        batch_op.drop_column('alias_domain')
